# Input / output, general run definitions
import sys
import os
# To not have numpy start parallelizing on its own
os.environ["OMP_NUM_THREADS"] = "1"
# Read external packages
import numpy as np
import copy as cp
import pymultinest
import json
import logging
from scipy.stats import binned_statistic
# Plotting
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator, LogLocator, NullFormatter
# Read own packages
from petitRADTRANS import Radtrans
from petitRADTRANS import nat_cst as nc
from .parameter import Parameter
from .data import Data
from .plotting import plot_specs,plot_data,contour_corner
from .rebin_give_width import rebin_give_width as rgw

class Retrieval:
    """
    This class implements the retrieval method using petitRADTRANS and pymultinest.
    A RetrievalConfig object is passed to this class to describe the retrieval data, parameters
    and priors. The run() method then uses pymultinest to sample the parameter space, producing
    posterior distributions for parameters and bayesian evidence for models.
    Various useful plotting functions have also been included, and can be run once the retrieval is
    complete.

    Args:
        run_definition : RetrievalConfig
            A RetrievalConfig object that describes the retrieval to be run. This is the user
            facing class that must be setup for every retrieval.
        output_dir : Str
            The directory in which the output folders should be written
        test_plotting : Bool
            Only use when running locally. A boolean flag that will produce plots
            for each sample when pymultinest is run.
        sample_spec : Bool
            Produce plots and data files for 100 randomly sampled outputs from pymultinest.
        sampling_efficiency : Float
            pymultinest sampling efficiency
        const_efficiency_mode : Bool
            pymultinest constant efficiency mode
        n_live_points : Int
            Number of live points to use in pymultinest.
        resume : bool
            Continue existing retrieval. If FALSE THIS WILL OVERWRITE YOUR EXISTING RETRIEVAL.
        bayes_factor_species : Str
            A pRT species that should be removed to test for the bayesian evidence for it's presence.
        corner_plot_names : List(Str)
            List of additional retrieval names that should be included in the corner plot.
        short_names : List(Str)
            For each corner_plot_name, a shorter name to be included when plotting.
        plot_multiple_retrieval_names : List(Str)
            List of additional retrievals to include when plotting spectra and PT profiles.
            Not yet implemented.
    """

    def __init__(self,
                 run_definition,
                 output_dir = "",
                 test_plotting = False,
                 sample_spec = False,
                 sampling_efficiency = 0.05,\
                 const_efficiency_mode = True, \
                 n_live_points = 4000,
                 resume = True,
                 bayes_factor_species = None,
                 corner_plot_names = None,
                 short_names = None,
                 pRT_plot_style = True,
                 plot_multiple_retrieval_names = None):
        self.rd = run_definition

        # Maybe inherit from retrieval config class?
        self.retrieval_name = self.rd.retrieval_name
        self.data = self.rd.data
        self.run_mode = self.rd.run_mode
        self.parameters = self.rd.parameters

        self.output_dir = output_dir
        if self.output_dir != "" and not self.output_dir.endswith("/"):
            self.output_dir += "/"

        self.remove_species = bayes_factor_species
        self.corner_files = corner_plot_names
        if self.corner_files is None:
            self.corner_files = [self.retrieval_name]
        self.short_names = short_names

        # Plotting variables
        self.best_fit_specs = {}
        self.best_fit_params = {}
        self.posterior_sample_specs = {}
        self.plotting = test_plotting
        self.PT_plot_mode = test_plotting
        self.evaluate_sample_spectra = sample_spec

        # Pymultinest stuff
        self.sampling_efficiency = sampling_efficiency
        self.const_efficiency_mode = const_efficiency_mode
        self.n_live_points = n_live_points
        self.resume = resume
        self.analyzer = None
        # TODO
        self.retrieval_list = plot_multiple_retrieval_names

        self.samples = {} #: The samples produced by pymultinest.
        self.param_dict = {}
        # Set up pretty plotting
        if pRT_plot_style:
            import petitRADTRANS.retrieval.plot_style
        # Path to input opacities
        self.path = os.environ.get("pRT_input_data_path")
        if self.path is None:
            print('Path to input data not specified!')
            print('Please set pRT_input_data_path variable in .bashrc/.bash_profile or specify path via')
            print('    import os')
            print('    os.environ["pRT_input_data_path"] = "absolute/path/of/the/folder/input_data"')
            print('before creating a Radtrans object or loading the nat_cst module.')
            logging.error("pRT_input_data_path not set")
            sys.exit(1)
        if not self.path.endswith("/"):
            self.path += "/"
        # Setup Directories
        if not os.path.isdir(self.output_dir + 'out_PMN/'):
            os.makedirs(self.output_dir + 'out_PMN', exist_ok=True)
        if not os.path.isdir(self.output_dir + 'evaluate_' + self.retrieval_name +'/'):
            os.makedirs(self.output_dir + 'evaluate_' + self.retrieval_name, exist_ok=True)

        # Setup pRT Objects for each data structure.
        print("Setting up PRT Objects")
        self.setup_data()
        self.generate_retrieval_summary()

    def run(self):
        """
        Run mode for the class. Uses pynultinest to sample parameter space
        and produce standard PMN outputs.
        """

        prefix = self.output_dir + 'out_PMN/'+self.retrieval_name+'_'

        if len(self.output_dir + 'out_PMN/') > 100:
            logging.error("PyMultinest requires output directory names to be <100 characters.")
            sys.exit(3)

        # How many free parameters?
        n_params = 0
        free_parameter_names = []
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                free_parameter_names.append(self.parameters[pp].name)
                n_params += 1
        if self.run_mode == 'retrieval':
            print("Starting retrieval: " + self.retrieval_name+'\n')
            json.dump(free_parameter_names, \
                    open(self.output_dir + 'out_PMN/'+self.retrieval_name+'_params.json', 'w'))
            pymultinest.run(self.log_likelihood,
                            self.prior,
                            n_params,
                            outputfiles_basename=prefix,
                            resume = self.resume,
                            verbose = True,
                            sampling_efficiency = self.sampling_efficiency,
                            const_efficiency_mode = self.const_efficiency_mode,
                            n_live_points = self.n_live_points)
        self.analyzer = pymultinest.Analyzer(n_params = n_params,
                                                outputfiles_basename = prefix)
        s = self.analyzer.get_stats()
        self.run_mode = 'evaluate'
        self.generate_retrieval_summary(s)
        json.dump(s, open(prefix + 'stats.json', 'w'), indent=4)
        print('  marginal likelihood:')
        print('    ln Z = %.1f +- %.1f' % (s['global evidence'], s['global evidence error']))
        print('  parameters:')


        for p, m in zip(free_parameter_names, s['marginals']):
            lo, hi = m['1sigma']
            med = m['median']
            sigma = (hi - lo) / 2
            if sigma == 0:
                i = 3
            else:
                i = max(0, int(-np.floor(np.log10(sigma))) + 1)
            fmt = '%%.%df' % i
            fmts = '\t'.join(['    %-15s' + fmt + " +- " + fmt])
            print(fmts % (p, med, sigma))

    def run_ultranest(self):
        """
        Run mode for the class. Uses ultranest to sample parameter space
        and produce standard outputs.
        """

        logging.warning("ultranest mode is still in development. Proceed with caution")
        try:
            import ultranest as un
        except ImportError:
            logging.error("Could not import ultranest. Exiting.")
            sys.exit(1)
        # TODO: autodetect PMN vs UN outputs
        prefix = self.output_dir + 'out_PMN/'+self.retrieval_name+'_'
        if self.run_mode == 'retrieval':
            print("Starting retrieval: " + self.retrieval_name+'\n')
            # How many free parameters?
            n_params = 0
            free_parameter_names = []
            for pp in self.parameters:
                if self.parameters[pp].is_free_parameter:
                    free_parameter_names.append(self.parameters[pp].name)
                    n_params += 1
            json.dump(free_parameter_names, \
                    open(prefix+'_params.json', 'w'))
            sampler = un.ReactiveNestedSampler(free_parameter_names,
                                               self.log_likelihood,
                                               self.prior,
                                               log_dir=self.output_dir + "out_" + self.retrieval_name,
                                               resume=self.resume)

            result = sampler.run()
            sampler.print_results()
        return result

    def generate_retrieval_summary(self,stats = None):
        """
        This function produces a human-readable text file describing the retrieval.
        It includes all of the fixed and free parameters, the limits of the priors (if uniform),
        a description of the data used, and if the retrieval is complete, a summary of the
        best fit parameters and model evidence.

        Args:
            stats : dict
                A Pymultinest stats dictionary, from Analyzer.get_stats().
                This contains the evidence and best fit parameters.
        """

        with open(self.output_dir + "evaluate_" + self.retrieval_name + "/" + self.retrieval_name +\
                  "_ret_summary.txt", "w+") as summary:
            from datetime import datetime
            summary.write(self.retrieval_name + '\n')
            summary.write(datetime.now().strftime("%Y-%m-%d, %H:%M:%S") + '\n')
            summary.write(self.output_dir + '\n\n')
            summary.write("Fixed Parameters\n")
            for key,value in self.parameters.items():
                if key in ['pressure_simple', 'pressure_width', 'pressure_scaling']:
                    continue
                if not value.is_free_parameter:
                    summary.write("    " + key + " = " + str(round(value.value,3)) + '\n')
            summary.write('\n')
            summary.write("Free Parameters, Prior^-1(0), Prior^-1(1)\n")
            for key,value in self.parameters.items():
                if value.is_free_parameter:
                    low = value.transform_prior_cube_coordinate(0.0000001)
                    high = value.transform_prior_cube_coordinate(0.9999999)
                    if value.corner_transform is not None:
                        low = value.corner_transform(low)
                        high = value.corner_transform(high)
                    summary.write("    " +key + " = " + str(round(low,3)) + ", " +\
                                  str(round(high,3)) + '\n')
            summary.write('\n')
            summary.write("Data\n")
            for name,dd in self.data.items():
                summary.write(name+'\n')
                summary.write("    " + dd.path_to_observations + '\n')
                if dd.model_generating_function is not None:
                    summary.write("    Model Function = " + dd.model_generating_function.__name__+ '\n')
                if dd.scale:
                    summary.write("    scale factor = " + str(round(dd.scale_factor,2))+ '\n')
                if dd.data_resolution is not None:
                    summary.write("    data resolution = " + str(int(dd.data_resolution))+ '\n')
                if dd.model_resolution is not None:
                    summary.write("    model resolution = " + str(int(dd.model_resolution))+ '\n')
                if dd.photometry:
                    summary.write("    photometric width = " + str(round(dd.photometry_range[0],4)) + \
                                  "--" + str(round(dd.photometry_range[1],4)) + " um"+ '\n')
                    summary.write("    Photometric transform function = " + \
                                  dd.photometric_transformation_function.__name__+ '\n')
            summary.write('\n')
            if stats is not None:
                summary.write("Multinest Outputs\n")
                summary.write('  marginal likelihood:\n')
                summary.write('    log Z = %.1f +- %.1f\n' % \
                             (stats['global evidence']/np.log(10), stats['global evidence error']/np.log(10)))
                summary.write('    ln Z = %.1f +- %.1f\n' % (stats['global evidence'], \
                              stats['global evidence error']))
                summary.write("  Statistical Fit Parameters\n")

                free_params = []
                for key,value in self.parameters.items():
                    if value.is_free_parameter:
                        free_params.append(key)

                for p, m in zip(free_params, stats['marginals']):
                    lo, hi = m['1sigma']
                    med = m['median']
                    sigma = (hi - lo) / 2
                    if sigma == 0:
                        i = 3
                    else:
                        i = max(0, int(-np.floor(np.log10(sigma))) + 1)
                    fmt = '%%.%df' % i
                    fmts = '\t'.join(['    %-15s' + fmt + " +- " + fmt])
                    summary.write(fmts % (p, med, sigma) + '\n')
                summary.write('\n')
            if self.run_mode == 'evaluate':
                summary.write("Best Fit Parameters\n")
                if not self.best_fit_params:
                    self.get_samples(self.output_dir)
                    samples_use = self.samples[self.retrieval_name]
                    parameters_read = self.param_dict[self.retrieval_name]
                    # Get best-fit index
                    logL = samples_use[:,-1]
                    best_fit_index = np.argmax(logL)
                    self.get_best_fit_params(samples_use[best_fit_index,:-1],parameters_read)
                for key,value in self.best_fit_params.items():
                    if key in ['pressure_simple', 'pressure_width', 'pressure_scaling']:
                        continue
                    out = value.value
                    if self.parameters[key].corner_transform is not None:
                        out = self.parameters[key].corner_transform(out)
                    fmt = '%.3f' % out
                    summary.write("    " +key + " = " + fmt + '\n')

    def setup_data(self,scaling=10,width = 3):
        """
        Creates a pRT object for each data set that asks for a unique object.
        Checks if there are low resolution c-k models from exo-k, and creates them if necessary.
        The scaling and width parameters adjust the AMR grid as described in RetrievalConfig.setup_pres
        and models.fixed_length_amr. It is recommended to keep the defaults.

        Args:
            scaling : int
                A multiplicative factor that determines the size of the full high resolution pressure grid,
                which will have length self.p_global.shape[0] * scaling.
            width : int
                The number of cells in the low pressure grid to replace with the high resolution grid.
        """

        for name,dd in self.data.items():
            # Only create if there's no other data
            # object using the same pRT object
            if dd.external_pRT_reference is None:
                if dd.opacity_mode == 'c-k' and dd.model_resolution is not None:
                    # Use ExoK to have low res models.
                    species = []
                    # Check if low res opacities already exist
                    for line in self.rd.line_species:
                        if not os.path.isdir(self.path + "opacities/lines/corr_k/" +\
                                                line + "_R_" + \
                                                str(dd.model_resolution)):
                            species.append(line)
                    # If not, setup low-res c-k tables
                    if len(species)>0:
                        from .util import getMM
                        # Automatically build the entire table
                        atmosphere = Radtrans(line_species = species,
                                                wlen_bords_micron = [0.1, 251.])
                        prt_path = self.path
                        ck_path = prt_path + 'opacities/lines/corr_k/'
                        print("Saving to " + ck_path)
                        print("Resolution: ", dd.model_resolution)
                        masses = {}
                        for spec in species:
                            masses[spec.split('_')[0]] = getMM(spec)
                        atmosphere.write_out_rebin(int(dd.model_resolution),
                                                    path = ck_path,
                                                    species = species,
                                                    masses = masses)
                    species = []
                    for spec in self.rd.line_species:
                        species.append(spec + "_R_" + str(dd.model_resolution))
                else:
                    # Otherwise for 'lbl' or no model_resolution binning,
                    # we just use the default species.
                    species = cp.copy(self.rd.line_species)
                lbl_samp = None
                if dd.opacity_mode == 'lbl' and dd.model_resolution is not None:
                    lbl_samp = int(1e6/dd.model_resolution)

                # Setup the pRT objects for the given dataset
                rt_object = Radtrans(line_species = cp.copy(species), \
                                    rayleigh_species= cp.copy(self.rd.rayleigh_species), \
                                    continuum_opacities = cp.copy(self.rd.continuum_opacities), \
                                    cloud_species = cp.copy(self.rd.cloud_species), \
                                    mode=dd.opacity_mode, \
                                    wlen_bords_micron = dd.wlen_range_pRT,
                                    do_scat_emis = self.rd.scattering,
                                    lbl_opacity_sampling = lbl_samp)

                # Create random P-T profile to create RT arrays of the Radtrans object.
                if self.rd.AMR:
                    p = self.rd._setup_pres(scaling,width)
                else:
                    p = self.rd.p_global
                rt_object.setup_opa_structure(p)
                dd.pRT_object = rt_object

    def prior(self, cube, ndim, nparams):
        """
        pyMultinest Prior function. Transforms unit hypercube into physical space.
        """

        i_p = 0
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                cube[i_p] = self.parameters[pp].get_param_uniform(cube[i_p])
                i_p += 1

    def log_likelihood(self,cube,ndim,nparam):
        """
        pyMultiNest required likelihood function.

        This function wraps the model computation and log-likelihood calculations
        for pyMultiNest to sample. If PT_plot_mode is True, it will return the
        calculate only the pressure and temperature arrays rather than the wavlength
        and flux. If run_mode is evaluate, it will save the provided sample to the
        best-fit spectrum file, and add it to the best_fit_specs dictionary.
        If evaluate_sample_spectra is true, it will store the spectrum in
        posterior_sample_specs.

        Args:
            cube : numpy.ndarray
                The transformed unit hypercube, providing the parameter values
                to be passed to the model_generating_function.
            ndim : int
                The number of dimensions of the problem
            nparam : int
                The number of parameters in the fit.

        Returns:
            log_likelihood : float
                The (negative) log likelihood of the model given the data.
        """

        log_likelihood = 0.
        log_prior      = 0.

        i_p = 0 # parameter count
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                self.parameters[pp].set_param(cube[i_p])
                i_p += 1

        for name,dd in self.data.items():
            # Only calculate spectra within a given
            # wlen range once
            if dd.scale:
                dd.scale_factor = self.parameters[name + "_scale_factor"].value
            if dd.external_pRT_reference is None:
                if not self.PT_plot_mode:
                    # Compute the model
                    wlen_model, spectrum_model = \
                        dd.model_generating_function(dd.pRT_object,
                                                    self.parameters,
                                                    self.PT_plot_mode,
                                                    AMR = self.rd.AMR)
                    # Sanity checks on outputs
                    #print(spectrum_model)
                    if spectrum_model is None:
                        return -np.inf
                    if np.isnan(spectrum_model).any():
                        return -np.inf
                    log_likelihood += dd.get_chisq(wlen_model,
                                            spectrum_model,
                                            self.plotting)
                else:
                    # Get the PT profile
                    if name == self.rd.plot_kwargs["take_PTs_from"]:
                        pressures, temperatures = \
                            dd.model_generating_function(dd.pRT_object,
                                                         self.parameters,
                                                         self.PT_plot_mode,
                                                         AMR = self.rd.AMR)
                        return pressures, temperatures
                # Save sampled outputs if necessary.
                if self.run_mode == 'evaluate':
                    if self.evaluate_sample_spectra:
                        self.posterior_sample_specs[name] = [wlen_model, \
                                                spectrum_model]
                    else:
                        np.savetxt(self.output_dir + 'evaluate_' + self.retrieval_name + \
                                   '/model_spec_best_fit_'+
                                   name.replace('/','_').replace('.','_')+'.dat',
                                   np.column_stack((wlen_model,
                                                    spectrum_model)))

                        self.best_fit_specs[name] = [wlen_model, \
                                                spectrum_model]

            # Check for data using the same pRT object,
            # calculate log_likelihood
            for de_name,dede in self.data.items():
                if dede.external_pRT_reference is not None:
                    if dede.scale:
                        dd.scale_factor = self.parameters[de_name + "_scale_factor"].value
                    if dede.external_pRT_reference == name:
                        log_likelihood += dede.get_chisq(wlen_model, \
                                        spectrum_model, \
                                        self.plotting)
        #print(log_likelihood)
        return log_likelihood + log_prior

    def get_samples(self, output_dir = None, ret_names = []):
        """
        This function looks in the given output directory and finds the post_equal_weights
        file associated with the current retrieval name.

        Args:
            output_dir : str
                Parent directory of the out_PMN/RETRIEVALNAME_post_equal_weights.dat file
            ret_names : List(str)
                A list of retrieval names to add to the sample and parameter dictionary.
                Functions the same as setting corner_files during initialisation.

        Returns:
            sample_dict : dict
                A dictionary with keys being the name of the retrieval, and values are a numpy
                ndarray containing the samples in the post_equal_weights file
            parameter_dict : dict
                A dictionary with keys being the name of the retrieval, and values are a list of names
                of the parameters used in the retrieval. The first name corresponds to the first column
                of the samples, and so on.
        """

        if output_dir is None:
            output_dir = self.output_dir
        for name in self.corner_files:
            samples = np.genfromtxt(output_dir +'out_PMN/'+ \
                                    name+ \
                                    '_post_equal_weights.dat')

            parameters_read = json.load(open(output_dir + 'out_PMN/'+ \
                                        name+ \
                                        '_params.json'))
            self.samples[name] = samples
            self.param_dict[name] = parameters_read
        for name in ret_names:
            samples = np.genfromtxt(output_dir +'out_PMN/'+ \
                                    name+ \
                                    '_post_equal_weights.dat')

            parameters_read = json.load(open(output_dir + 'out_PMN/'+ \
                                        name+ \
                                        '_params.json'))
            self.samples[name] = samples
            self.param_dict[name] = parameters_read
        return self.samples, self.param_dict

    def get_best_fit_params(self,best_fit_params,parameters_read):
        """
        This function converts the sample from the post_equal_weights file with the maximum
        log likelihood, and converts it into a dictionary of Parameters that can be used in
        a model function.

        Args:
            best_fit_params : numpy.ndarray
                An array of the best fit parameter values (or any other sample)
            parameters_read : list
                A list of the free parameters as read from the output files.
        """

        i_p = 0
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                for i_s in range(len(parameters_read)):
                    if parameters_read[i_s] == self.parameters[pp].name:
                        self.best_fit_params[self.parameters[pp].name] = \
                            Parameter(pp,False,value=best_fit_params[i_p])
                        i_p += 1
            else:
                self.best_fit_params[pp] = Parameter(pp,False,value=self.parameters[pp].value)
        return self.best_fit_params

    def get_best_fit_model(self,best_fit_params,parameters_read,model_generating_func = None,ret_name = None):
        """
        This function uses the best fit parameters to generate a pRT model that spans the entire wavelength
        range of the retrieval, to be used in plots.

        Args:
            best_fit_params : numpy.ndarray
                A numpy array containing the best fit parameters, to be passed to get_best_fit_params
            parameters_read : list
                A list of the free parameters as read from the output files.
            model_generating_fun : method
                A function that will take in the standard 'model' arguments
                (pRT_object, params, pt_plot_mode, AMR, resolution)
                and will return the wavlength and flux arrays as calculated by petitRadTrans.
                If no argument is given, it uses the method of the dataset given in the take_PTs_from kwarg.
            ret_name : str
                If plotting a fit from a different retrieval, input the retrieval name to be included.

        Returns:
            bf_wlen : numpy.ndarray
                The wavelength array of the best fit model
            bf_spectrum : numpy.ndarray
                The emission or transmission spectrum array, with the same shape as bf_wlen
        """

        print("Computing Best Fit Model, this may take a minute...")
        if ret_name is None:
            ret_name = self.retrieval_name

        # Find the boundaries of the wavelength range to calculate
        wmin = 99999.0
        wmax = 0.0
        for name,dd in self.data.items():
            if dd.wlen_range_pRT[0] < wmin:
                wmin = dd.wlen_range_pRT[0]
            if dd.wlen_range_pRT[1] > wmax:
                wmax = dd.wlen_range_pRT[1]
        # Set up parameter dictionary
        if not self.retrieval_name in self.best_fit_specs.keys():
            self.get_best_fit_params(best_fit_params,parameters_read)

        # Setup the pRT object
        bf_prt = Radtrans(line_species = cp.copy(self.rd.line_species), \
                            rayleigh_species= cp.copy(self.rd.rayleigh_species), \
                            continuum_opacities = cp.copy(self.rd.continuum_opacities), \
                            cloud_species = cp.copy(self.rd.cloud_species), \
                            mode='c-k', \
                            wlen_bords_micron = [wmin*0.98,wmax*1.02],
                            do_scat_emis = self.rd.scattering)
        if self.rd.AMR:
            p = self.rd._setup_pres()
        else:
            p = self.rd.p_global
        bf_prt.setup_opa_structure(p)

        # Check what model function we're using
        if model_generating_func is None:
            mg_func = self.data[self.rd.plot_kwargs["take_PTs_from"]].model_generating_function
        else:
            mg_func = model_generating_func

        # get the spectrum
        bf_wlen, bf_spectrum= mg_func(bf_prt,
                                      self.best_fit_params,
                                      PT_plot_mode= False,
                                      AMR = self.rd.AMR)
        # Add to the dictionary.
        self.best_fit_specs[ret_name]= [bf_wlen,bf_spectrum]
        np.save(self.output_dir + "evaluate_" + \
                self.retrieval_name + "/" + \
                ret_name + "best_fit_model_full",
                np.column_stack([bf_wlen,bf_spectrum]))
        return bf_wlen, bf_spectrum

    def get_abundances(self,sample):
        """
        This function returns the abundances of each species as a function of pressure

        Args:
            sample : numpy.ndarray
                A sample from the pymultinest output, the abundances returned will be
                computed for this set of parameters.
        Returns:
            abundances : dict
                A dictionary of abundances. The keys are the species name,
                the values are the mass fraction abundances at each pressure
            MMW : numpy.ndarray
                The mean molecular weight at each pressure level in the atmosphere.
        """
        params = self.get_best_fit_params(sample)
        from petitRADTRANS.retrieval.models import get_abundances
        name = self.rd.plot_kwargs["take_PTs_from"]
        abundances, MMW, _ = get_abundances(self.rd.p_global,
                                            self.data[name].pRT_object.temperature,
                                            self.data[name].pRT_object.line_species,
                                            self.data[name].pRT_object.cloud_species,
                                            params,
                                            AMR=False)
        return abundances, MMW

    def get_evidence(self,ret_name = ""):
        """
        Get the log10 Z and error for the retrieval

        This function uses the pymultinest analyzer to
        get the evidence for the current retrieval_name
        by default, though any retrieval_name in the
        out_PMN folder can be passed as an argument -
        useful for when you're comparing multiple similar
        models. This value is also printed in the summary file.

        Args:
            ret_name : string
                The name of the retrieval that prepends all of the PMN
                output files.
        """
        analyzer = self.get_analyzer(ret_name)
        s = analyzer.get_stats()
        return s['global evidence']/np.log(10), s['global evidence error']/np.log(10)

    def get_analyzer(self,ret_name = ""):
        """
        Get the PMN analyer from a retrieval run

        This function uses gets the PMN analyzer object
        for the current retrieval_name by default,
        though any retrieval_name in the out_PMN folder can
        be passed as an argument - useful for when you're
        comparing multiple similar models.

        Args:
            ret_name : string
                The name of the retrieval that prepends all of the PMN
                output files.
        """
        # Avoid loading if we just want the current retrievals output
        if ret_name is "" and self.analyzer is not None:
            return self.analyzer
        if ret_name is "":
            ret_name = self.retrieval_name
        prefix = self.output_dir + 'out_PMN/'+ret_name+'_'

        # How many free parameters?
        n_params = 0
        free_parameter_names = []
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                free_parameter_names.append(self.parameters[pp].name)
                n_params += 1

        # Get the outputs
        analyzer = pymultinest.Analyzer(n_params = n_params,
                                        outputfiles_basename = prefix)
        if ret_name == self.retrieval_name:
            self.analyzer = analyzer
        return analyzer
#############################################################
# Plotting functions
#############################################################
    def plot_all(self, output_dir = None):
        """
        Produces plots for the best fit spectrum, a sample of 100 output spectra,
        the best fit PT profile and a corner plot for parameters specified in the
        run definition.
        """

        if not self.run_mode == 'evaluate':
            logging.warning("Not in evaluate mode. Changing run mode to evaluate.")
            self.run_mode = 'evaluate'
        if output_dir is None:
            output_dir = self.output_dir
        sample_dict, parameter_dict = self.get_samples(output_dir)

        ###########################################
        # Plot best-fit spectrum
        ###########################################
        samples_use = cp.copy(sample_dict[self.retrieval_name])
        parameters_read = cp.copy(parameter_dict[self.retrieval_name])
        i_p = 0
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                for i_s in range(len(parameters_read)):
                    if parameters_read[i_s] == self.parameters[pp].name:
                        samples_use[:,i_p] = sample_dict[self.retrieval_name][:, i_s]
                i_p += 1

        print("Best fit parameters")
        i_p = 0
        # Get best-fit index
        logL = samples_use[:,-1]
        best_fit_index = np.argmax(logL)
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                for i_s in range(len(parameters_read)):
                    if parameters_read[i_s] == self.parameters[pp].name:
                        print(self.parameters[pp].name, samples_use[best_fit_index][i_p])
                        i_p += 1

        # Plotting
        self.plot_spectra(samples_use,parameters_read)
        self.plot_sampled(samples_use, parameters_read)
        self.plot_PT(sample_dict,parameters_read)
        self.plot_corner(sample_dict,parameter_dict,parameters_read)
        print("Done!")
        return

    def plot_spectra(self,samples_use,parameters_read,model_generating_func = None):
        """
        Plot the best fit spectrum, the data from each dataset and the residuals between the two.
        Saves a file to OUTPUT_DIR/evaluate_RETRIEVAL_NAME/best_fit_spec.pdf

        Args:
            samples_use : numpy.ndarray
                An array of the samples from the post_equal_weights file, used to find the best fit sample
            parameters_read : list
                A list of the free parameters as read from the output files.
            model_generating_fun : method
                A function that will take in the standard 'model' arguments
                (pRT_object, params, pt_plot_mode, AMR, resolution)
                and will return the wavlength and flux arrays as calculated by petitRadTrans.
                If no argument is given, it uses the method of the first dataset included in the retrieval.

        Returns:
            fig : matplotlib.figure
                The matplotlib figure, containing the data, best fit spectrum and residuals.
            ax : matplotlib.axes
                The upper pane of the plot, containing the best fit spectrum and data
            ax_r : matplotlib.axes
                The lower pane of the plot, containing the residuals between the fit and the data
        """

        self.evaluate_sample_spectra = False
        #TODO: include plotting of multiple retrievals
        if not self.run_mode == 'evaluate':
            logging.warning("Not in evaluate mode. Changing run mode to evaluate.")
            self.run_mode = 'evaluate'
        print("Plotting Best-fit spectrum")
        fig, axes = fig, axes = plt.subplots(nrows=2, ncols=1, sharex='col', sharey=False,
                               gridspec_kw={'height_ratios': [2.5, 1],'hspace':0.1},
                               figsize=(20, 10))
        ax = axes[0] # Normal Spectrum axis
        ax_r = axes[1] # residual axis

        # Get best-fit index
        logL = samples_use[:,-1]
        best_fit_index = np.argmax(logL)

        # Setup best fit spectrum
        # First get the fit for each dataset for the residual plots
        self.log_likelihood(samples_use[best_fit_index, :-1], 0, 0)
        # Then get the full wavelength range
        bf_wlen, bf_spectrum = self.get_best_fit_model(samples_use[best_fit_index, :-1],\
                                                       parameters_read,model_generating_func)

        # Iterate through each dataset, plotting the data and the residuals.
        for name,dd in self.data.items():
            # If the user has specified a resolution, rebin to that
            if not dd.photometry:
                try:
                    # Sometimes this fails, I'm not super sure why.
                    resolution_data = np.mean(dd.wlen[1:]/np.diff(dd.wlen))
                    ratio = resolution_data / self.rd.plot_kwargs["resolution"]
                    if int(ratio) > 1:
                        flux,edges,_ = binned_statistic(dd.wlen,dd.flux,'mean',dd.wlen.shape[0]/ratio)
                        error,_,_ = binned_statistic(dd.wlen,dd.flux_error,\
                                                    'mean',dd.wlen.shape[0]/ratio)/np.sqrt(ratio)
                        wlen = np.array([(edges[i]+edges[i+1])/2.0 for i in range(edges.shape[0]-1)])
                        # Old method
                        #wlen = nc.running_mean(dd.wlen, int(ratio))[::int(ratio)]
                        #error = nc.running_mean(dd.flux_error / int(np.sqrt(ratio)), \
                        #                        int(ratio))[::int(ratio)]
                        #flux = nc.running_mean(dd.flux, \
                        #                        int(ratio))[::int(ratio)]
                    else:
                        wlen = dd.wlen
                        error = dd.flux_error
                        flux = dd.flux
                except:
                    wlen = dd.wlen
                    error = dd.flux_error
                    flux = dd.flux
                # Setup bins to rebin the best fit model to find the residuals
                wlen_bins = np.zeros_like(wlen)
                wlen_bins[:-1] = np.diff(wlen)
                wlen_bins[-1] = wlen_bins[-2]
            else:
                wlen = np.mean(dd.width_photometry)
                flux = dd.flux
                error = dd.flux_error
                wlen_bins = dd.wlen_bins

            # If the data has an arbitrary retrieved scaling factor
            scale = dd.scale_factor

            if not dd.photometry:
                if dd.external_pRT_reference is None:
                    best_fit_binned = rgw(self.best_fit_specs[name][0], \
                                            self.best_fit_specs[name][1], \
                                            wlen, \
                                            wlen_bins)
                else:
                    best_fit_binned = rgw(self.best_fit_specs[dd.external_pRT_reference][0], \
                                self.best_fit_specs[dd.external_pRT_reference][1], \
                                wlen, \
                                wlen_bins)
            else:
                if dd.external_pRT_reference is None:
                    best_fit_binned = dd.photometric_transformation_function(self.best_fit_specs[name][0],
                                                                         self.best_fit_specs[name][1])
                    # Species functions give tuples of (flux,error)
                    try:
                        best_fit_binned = best_fit_binned[0]
                    except:
                        pass

                else:
                    best_fit_binned = \
                        dd.photometric_transformation_function(self.best_fit_specs[dd.external_pRT_reference][0],
                                                               self.best_fit_specs[dd.external_pRT_reference][1])
                    try:
                        best_fit_binned = best_fit_binned[0]
                    except:
                        pass
            # Plot the data
            marker = 'o'
            if dd.photometry:
                marker = 's'
            if not dd.photometry:
                ax.errorbar(wlen, \
                            flux * self.rd.plot_kwargs["y_axis_scaling"] * scale, \
                            yerr = error * self.rd.plot_kwargs["y_axis_scaling"] *scale, \
                            marker=marker, markeredgecolor='k', linewidth = 0, elinewidth = 2, \
                            label = dd.name, zorder =10, alpha = 0.9,)
            else:
                # Don't label photometry?
                ax.errorbar(wlen, \
                            flux * self.rd.plot_kwargs["y_axis_scaling"] * scale, \
                            yerr = error * self.rd.plot_kwargs["y_axis_scaling"] *scale, \
                            xerr = dd.wlen_bins/2., linewidth = 0, elinewidth = 2, \
                            marker=marker, markeredgecolor='k', color = 'grey', zorder = 10, \
                            label = None, alpha = 0.6)
            # Plot the residuals
            col = ax.get_lines()[-1].get_color()
            if dd.external_pRT_reference is None:

                ax_r.errorbar(wlen, \
                            ((flux*scale) - best_fit_binned )/error ,
                            yerr = error/error,
                            color = col,
                            linewidth = 0, elinewidth = 2, \
                            marker=marker, markeredgecolor='k', zorder = 10,
                            alpha = 0.9)
            else:
                ax_r.errorbar(wlen, \
                        ((flux*scale) - best_fit_binned )/error,
                        yerr = error/error,
                        color = col,
                        linewidth = 0, elinewidth = 2, \
                        marker=marker, markeredgecolor='k', zorder = 10,
                        alpha = 0.9)
        # Plot the best fit model
        ax.plot(bf_wlen, \
                bf_spectrum * self.rd.plot_kwargs["y_axis_scaling"],
                label = 'Best Fit Model',
                linewidth=4,
                alpha = 0.5,
                color = 'r')
        # Plot the shading in the residual plot
        yabs_max = abs(max(ax_r.get_ylim(), key=abs))
        lims = ax.get_xlim()
        lim_y = ax.get_ylim()
        lim_y = [lim_y[0],lim_y[1]*1.12]
        ax.set_ylim(lim_y)
        # weird scaling to get axis to look ok on log plots
        if self.rd.plot_kwargs["xscale"] == 'log':
            lims = [lims[0]*1.09,lims[1]*1.02]
        else:
            lims = [bf_wlen[0]*0.98,bf_wlen[-1]*1.02]
        ax.set_xlim(lims)
        ax_r.set_xlim(lims)
        ax_r.set_ylim(ymin=-yabs_max, ymax=yabs_max)
        ax_r.fill_between(lims,-1,1,color='dimgrey',alpha=0.4,zorder = -10)
        ax_r.fill_between(lims,-3,3,color='darkgrey',alpha=0.3,zorder = -9)
        ax_r.fill_between(lims,-5,5,color='lightgrey',alpha=0.3,zorder = -8)
        ax_r.axhline(linestyle = '--', color = 'k',alpha=0.8, linewidth=2)

        # Making the plots pretty
        try:
            ax.set_xscale(self.rd.plot_kwargs["xscale"])
        except:
            pass
        try:
            ax.set_yscale(self.rd.plot_kwargs["yscale"])
        except:
            pass

        # Fancy ticks for upper pane
        ax.tick_params(axis="both",direction="in",length=10,bottom=True, top=True, left=True, right=True)
        try:
            ax.xaxis.set_major_formatter('{x:.1f}')
        except:
            logging.warning("Please update to matplotlib 3.3.4 or greater")
            pass

        if self.rd.plot_kwargs["xscale"] == 'log':
            # For the minor ticks, use no labels; default NullFormatter.
            x_major = LogLocator(base = 10.0, subs = (1,2,3,4), numticks = 4)
            ax.xaxis.set_major_locator(x_major)
            x_minor = LogLocator(base = 10.0, subs = np.arange(0.1,10.1,0.1)*0.1, numticks = 100)
            ax.xaxis.set_minor_locator(x_minor)
            ax.xaxis.set_minor_formatter(NullFormatter())
        else:
            ax.xaxis.set_minor_locator(AutoMinorLocator())
            ax.tick_params(axis='both', which='minor',
                           bottom=True, top=True, left=True, right=True,
                           direction='in',length=5)
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        ax.tick_params(axis='both', which='minor',
                       bottom=True, top=True, left=True, right=True,
                       direction='in',length=5)
        ax.set_ylabel(self.rd.plot_kwargs["spec_ylabel"])

        # Fancy ticks for lower pane
        ax_r.tick_params(axis="both",direction="in",length=10,bottom=True, top=True, left=True, right=True)

        try:
            ax_r.xaxis.set_major_formatter('{x:.1f}')
        except:
            logging.warning("Please update to matplotlib 3.3.4 or greater")
            pass

        if self.rd.plot_kwargs["xscale"] == 'log':
            # For the minor ticks, use no labels; default NullFormatter.
            x_major = LogLocator(base = 10.0, subs = (1,2,3,4), numticks = 4)
            ax_r.xaxis.set_major_locator(x_major)
            x_minor = LogLocator(base = 10.0, subs = np.arange(0.1,10.1,0.1)*0.1, numticks = 100)
            ax_r.xaxis.set_minor_locator(x_minor)
            ax_r.xaxis.set_minor_formatter(NullFormatter())
        else:
            ax_r.xaxis.set_minor_locator(AutoMinorLocator())
            ax_r.tick_params(axis='both', which='minor',
                             bottom=True, top=True, left=True, right=True,
                             direction='in',length=5)
        ax_r.yaxis.set_minor_locator(AutoMinorLocator())
        ax_r.tick_params(axis='both', which='minor',
                         bottom=True, top=True, left=True, right=True,
                         direction='in',length=5)
        ax_r.set_ylabel(r"Residuals [$\sigma$]")
        ax_r.set_xlabel(self.rd.plot_kwargs["spec_xlabel"])
        ax.legend(loc='upper center',ncol = len(self.data.keys())+1).set_zorder(1002)
        plt.tight_layout()
        plt.savefig(self.output_dir + 'evaluate_'+self.rd.retrieval_name +'/best_fit_spec.pdf')
        return fig, ax, ax_r

    def plot_sampled(self,samples_use,parameters_read, downsample_factor = None):
        """
        Plot a set of randomly sampled output spectra for each dataset in
        the retrieval.

        This will save nsample files for each dataset included in the retrieval.
        Note that if you change the model_resolution of your Data and rerun this
        function, the files will NOT be updated - if the files exists the function
        defaults to reading from file rather than recomputing. Delete all of the
        sample functions and run it again.

        Args:
            samples_use : np.ndarray
                posterior samples from pynmultinest outputs (post_equal_weights)
            downsample_factor : int
                Factor by which to reduce the resolution of the sampled model,
                for smoother plotting. Defaults to None. A value of None will result
                in the full resolution spectrum. Note that this factor can only
                reduce the resolution from the underlying model_resolution of the
                data.
        """

        self.evaluate_sample_spectra = True
        if not self.run_mode == 'evaluate':
            logging.warning("Not in evaluate mode. Changing run mode to evaluate.")
            self.run_mode = 'evaluate'
        self.rd.plot_kwargs["nsample"] = int(self.rd.plot_kwargs["nsample"])

        print("Plotting Best-fit spectrum with "+ str(self.rd.plot_kwargs["nsample"]) + " samples.")
        print("This could take some time...")
        len_samples = samples_use.shape[0]
        path = self.output_dir + 'evaluate_'+self.retrieval_name + "/"

        data_use= {}
        for name, dd in self.data.items():
            nsample_name = str(int(self.rd.plot_kwargs["nsample"])).zfill(int(np.log10(self.rd.plot_kwargs["nsample"])+1))
            if not os.path.exists(path + name.replace(' ','_').replace('/','_')+'_sampled_'+ nsample_name +'.dat'):
                data_use[name] = dd
        # TODO: doesn't guarantee same samples will be saved to file!
        fig,ax = plt.subplots(figsize = (16,10))
        for i_sample in range(int(self.rd.plot_kwargs["nsample"])):
            random_index = int(np.random.uniform()*len_samples)
            if data_use:
                self.log_likelihood(samples_use[random_index, :-1], 0, 0)
            for name,dd in data_use.items():
                if dd.external_pRT_reference is None:
                    np.savetxt(path +name.replace(' ','_').replace('/','_')+'_sampled_'+
                                str(int(i_sample+1)).zfill(5)+'.dat',
                                np.column_stack((self.posterior_sample_specs[name][0],
                                                 self.posterior_sample_specs[name][1])))
        # TODO: option for plotting of full bf model rather than by dataset
        for name,dd in self.data.items():
            fig, ax = plot_specs(fig,ax,path, name.replace(' ','_').replace('/','_'),
                                self.rd.plot_kwargs["nsample"],
                                '#ff9f9f', '#ff3d3d',
                                0, rebin_val = downsample_factor)

        for name,dd in self.data.items():
            fig, ax = plot_data(fig,ax,dd,
                                resolution = self.rd.plot_kwargs["resolution"],
                                scaling = self.rd.plot_kwargs["y_axis_scaling"])
            #plt.ylim([0.006, 0.0085])

        ax.set_xlabel('Wavelength [micron]')
        ax.set_ylabel(self.rd.plot_kwargs["spec_ylabel"])
        ax.legend(loc='best')
        plt.tight_layout()
        plt.savefig(path +'sampled_data.pdf',bbox_inches = 0.)
        self.evaluate_sample_spectra = False
        return fig, ax

    def plot_PT(self,sample_dict,parameters_read):
        """
        Plot the PT profile with error contours

        Args:
            samples_use : np.ndarray
                posterior samples from pynmultinest outputs (post_equal_weights)
            parameters_read : List
                Used to plot correct parameters, as some in self.parameters are not free, and
                aren't included in the PMN outputs

        Returns:
            fig : matplotlib.figure
            ax : matplotlib.axes
        """

        print("Plotting PT profiles")
        if not self.run_mode == 'evaluate':
            logging.warning("Not in evaluate mode. Changing run mode to evaluate.")
            self.run_mode = 'evaluate'
        self.PT_plot_mode = True
        samples_use = cp.copy(sample_dict[self.retrieval_name])
        i_p = 0
        for pp in self.parameters:
            if self.parameters[pp].is_free_parameter:
                for i_s in range(len(parameters_read)):
                    if parameters_read[i_s] == self.parameters[pp].name:
                        samples_use[:,i_p] = sample_dict[self.retrieval_name][:, i_s]
                i_p += 1

        temps = []
        for i_s in range(len(samples_use)):
            pressures, t = self.log_likelihood(samples_use[i_s, :-1], 0, 0)
            temps.append(t)

        temps = np.array(temps)
        temps_sort = np.sort(temps, axis=0)
        fig,ax = plt.subplots(figsize=(16, 10))
        len_samp = len(samples_use)
        ax.fill_betweenx(pressures, \
                        x1 = temps_sort[0, :], \
                        x2 = temps_sort[-1, :], \
                        color = 'cyan', label = 'all')
        ax.fill_betweenx(pressures, \
                        x1 = temps_sort[int(len_samp*(0.5-0.997/2.)), :], \
                        x2 = temps_sort[int(len_samp*(0.5+0.997/2.)), :], \
                        color = 'brown', label = '3 sig')
        ax.fill_betweenx(pressures, \
                        x1 = temps_sort[int(len_samp*(0.5-0.95/2.)), :], \
                        x2 = temps_sort[int(len_samp*(0.5+0.95/2.)), :], \
                        color = 'orange', label = '2 sig')
        ax.fill_betweenx(pressures, \
                        x1 = temps_sort[int(len_samp*(0.5-0.68/2.)), :], \
                        x2 = temps_sort[int(len_samp*(0.5+0.68/2.)), :], \
                        color = 'red', label = '1 sig')

        ax.set_yscale('log')
        try:
            ax.set_ylim(self.rd.plot_kwargs["press_limits"])
        except:
            ax.set_ylim([pressures[-1], pressures[0]])
        try:
            ax.set_xlim(self.rd.plot_kwargs["temp_limits"])
        except:
            pass
        ax.set_xlabel('Temperature [K]')
        ax.set_ylabel('Pressure [bar]')
        ax.legend(loc='best')
        plt.savefig(self.output_dir + 'evaluate_'+self.retrieval_name +'/PT_envelopes.pdf')
        return fig, ax

    def plot_corner(self,sample_dict,parameter_dict,parameters_read):
        """
        Make the corner plots

        Args:
            samples_dict : Dict
                Dictionary of samples from PMN outputs, with keys being retrieval names
            parameter_dict : Dict
                Dictionary of parameters for each of the retrievals to be plotted.
            parameters_read : List
                Used to plot correct parameters, as some in self.parameters are not free, and
                aren't included in the PMN outputs
        """

        if not self.run_mode == 'evaluate':
            logging.warning("Not in evaluate mode. Changing run mode to evaluate.")
            self.run_mode = 'evaluate'
        print("Making corner plot")
        sample_use_dict = {}
        p_plot_inds = {}
        p_ranges = {}
        p_use_dict = {}
        for name,params in parameter_dict.items():
            samples_use = cp.copy(sample_dict[name])
            parameters_use = cp.copy(params)
            parameter_plot_indices = []
            parameter_ranges       = []
            i_p = 0
            for pp in parameters_read:
                parameter_ranges.append(self.parameters[pp].corner_ranges)
                if self.parameters[pp].plot_in_corner:
                    parameter_plot_indices.append(i_p)
                if self.parameters[pp].corner_label is not None:
                    parameters_use[i_p] = self.parameters[pp].corner_label
                if self.parameters[pp].corner_transform is not None:
                    samples_use[:, i_p] = \
                        self.parameters[pp].corner_transform(samples_use[:, i_p])
                i_p += 1
            p_plot_inds[name] = parameter_plot_indices
            p_ranges[name] = parameter_ranges
            p_use_dict[name] = parameters_use
            sample_use_dict[name] = samples_use


        output_file = self.output_dir + 'evaluate_'+self.retrieval_name +'/corner_nice.pdf'


        # from Plotting
        contour_corner(sample_use_dict, \
                        p_use_dict, \
                        output_file, \
                        parameter_plot_indices = p_plot_inds,
                        parameter_ranges = p_ranges, \
                        true_values = None)



import sys
import os
import copy as cp
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
from scipy.interpolate import interp1d,CubicSpline
from petitRADTRANS import nat_cst as nc
from petitRADTRANS.retrieval import cloud_cond as fc
from .util import surf_to_meas, calc_MMW
"""
Models Module

This module contains a set of functions that generate the spectra used
in the petitRADTRANS retrieval. This includes setting up the
pressure-temperature structure, the chemistry, and the radiative
transfer to compute the emission or transmission spectrum.

All models must take the same set of inputs:

    pRT_object : petitRADTRANS.RadTrans
        This is the pRT object that is used to compute the spectrum
        It must be fully initialized prior to be used in the model function
    parameters : dict
        A dictionary of Parameter objects. The naming of the parameters
        must be consistent between the Priors and the model function you
        are using.
    PT_plot_mode : bool
        If this argument is True, the model function should return the pressure
        and temperature arrays before computing the flux.
    AMR : bool
        If this parameter is True, your model should allow for reshaping of the
        pressure and temperature arrays based on the position of the clouds or
        the location of the photosphere, increasing the resolution where required.
        For example, using the fixed_length_amr function defined below.
"""
# Global constants to reduce calculations and initializations.
PGLOBAL = np.logspace(-6,3,1000)

def emission_model_diseq(pRT_object,
                         parameters,
                         PT_plot_mode = False,
                         AMR = True):
    """
    Disequilibrium Chemistry Emission Model

    This model computes an emission spectrum based on disequilibrium carbon chemistry,
    equilibrium clouds and a spline temperature-pressure profile. (Molliere 2020).

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                *  log_g : Log of surface gravity
                *  R_pl : planet radius [cm]
                *  T_int : Interior temperature of the planet [K]
                *  T3 : Innermost temperature spline [K]
                *  T2 : Middle temperature spline [K]
                *  T1 : Outer temperature spline [K]
                *  alpha : power law index in tau = delta * press_cgs**alpha
                *  log_delta : proportionality factor in tau = delta * press_cgs**alpha
                *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                *  Fe/H : Metallicity
                *  C/O : Carbon to oxygen ratio
                *  log_kzz : Vertical mixing parameter
                *  fsed : sedimentation parameter
                *  log_X_cb : Scaling factor for equilibrium cloud abundances.
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed emission spectrum [W/m2/micron]
    """

    pglobal_check(pRT_object.press,
                    parameters['pressure_simple'].value,
                    parameters['pressure_scaling'].value)
    #for key, val in parameters.items():
    #    print(key,val.value)

    # Priors for these parameters are implemented here, as they depend on each other
    T3 = ((3./4.*parameters['T_int'].value**4.*(0.1+2./3.))**0.25)*(1-parameters['T3'].value)
    T2 = T3*(1.0-parameters['T2'].value)
    T1 = T2*(1.0-parameters['T1'].value)
    delta = ((10**(-3.0+5.0*parameters['log_delta'].value))*1e6)**(-parameters['alpha'].value)

    # Make the P-T profile
    temp_arr = np.array([T1,T2,T3])
    temperatures = PT_ret_model(temp_arr, \
                            delta,
                            parameters['alpha'].value,
                            parameters['T_int'].value,
                            PGLOBAL,
                            parameters['Fe/H'].value,
                            parameters['C/O'].value,
                            conv=True)
    if PT_plot_mode:
        return PGLOBAL, temperatures

    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index = get_abundances(PGLOBAL,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    Kzz_use = (10**parameters['log_kzz'].value ) * np.ones_like(PGLOBAL)

    # Only include the high resolution pressure array near the cloud base.
    pressures = PGLOBAL
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        Kzz_use = Kzz_use[small_index]
        #pRT_object.setup_opa_structure(pressures)
    # Calculate the spectrum
    if pressures.shape[0] != pRT_object.press.shape[0]:
        return None,None
    pRT_object.press = pressures*1e6

    gravity = 2.0
    R_pl = 1.0
    if 'log_g' in parameters.keys() and 'mass' in parameters.keys():
        gravity = 10**parameters['log_g'].value
        R_pl = np.sqrt(nc.G*parameters['mass'].value/gravity)
    elif 'log_g' in parameters.keys():
        gravity= 10**parameters['log_g'].value
        R_pl = parameters['R_pl'].value
    elif 'mass' in parameters.keys():
        R_pl = parameters['R_pl'].value
        gravity = nc.G * parameters['mass'].value/R_pl**2
    else:
        print("Pick two of log_g, R_pl and mass priors!")
        sys.exit(5)
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = False,
                        fsed = parameters['fsed'].value,
                        Kzz = Kzz_use,
                        sigma_lnorm = parameters['sigma_lnorm'].value)
    # Getting the model into correct units (W/m2/micron)
    wlen_model = nc.c/pRT_object.freq/1e-4
    wlen = nc.c/pRT_object.freq
    f_lambda = pRT_object.flux*nc.c/wlen**2.
    # convert to flux per m^2 (from flux per cm^2) cancels with step below
    #f_lambda = f_lambda * 1e4
    # convert to flux per micron (from flux per cm) cancels with step above
    #f_lambda = f_lambda * 1e-4
    # convert from ergs to Joule
    f_lambda = f_lambda * 1e-7
    spectrum_model = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)    #print(wlen_model,spectrum_model)
    return wlen_model, spectrum_model

def guillot_free_emission(pRT_object, \
                            parameters, \
                            PT_plot_mode = False,
                            AMR = False):
    """
    Free Chemistry Emission Model

    This model computes an emission spectrum based on free retrieval chemistry,
    free Ackermann-Marley clouds and a Guillot temperature-pressure profile. (Molliere 2018).

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                *  log_g : Log of surface gravity
                *  R_pl : planet radius [cm]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  gamma : Guillot gamma parameter
                *  log_kappa_IR : The log of the ratio between the infrared and optical opacities
                *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                *  log_kzz : Vertical mixing parameter
                *  fsed : sedimentation parameter
                *  species : Log abundances for each species in rd.line_list
                   (species stands in for the actual name)
                *  log_X_cb : Log cloud abundances.
                *  Pbase : log of cloud base pressure for each species.
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed emission spectrum [W/m2/micron]
    """

    #for key, val in parameters.items():
    #    print(key,val.value)

    pglobal_check(pRT_object.press,
                  parameters['pressure_simple'].value,
                  parameters['pressure_scaling'].value)

    temperatures = nc.guillot_global(PGLOBAL, \
                                10**parameters['log_kappa_IR'].value,
                                parameters['gamma'].value, \
                                10**parameters['log_g'].value, \
                                parameters['T_int'].value, \
                                parameters['T_equ'].value)
    Pbases = {}
    # TODO - identify species rather than hard coding
    Pbases['Fe(c)'] = fc.simple_cdf_Fe_free(PGLOBAL, temperatures,
                                10**parameters['log_X_cb_Fe(c)'].value)
    Pbases['MgSiO3(c)'] = fc.simple_cdf_MgSiO3_free(PGLOBAL, temperatures,
                                10**parameters['log_X_cb_MgSiO3(c)'].value)

    if AMR:
        p_clouds = np.array(list(Pbases.values()))
        pressures,small_index = fixed_length_amr(p_clouds,
                                                 PGLOBAL,
                                                 parameters['pressure_scaling'].value,
                                                 parameters['pressure_width'].value)
        pRT_object.press = pressures * 1e6
        temperatures = temperatures[small_index]
    else:
        pressures = pRT_object.press/1e6

    # If in evaluation mode, and PTs are supposed to be plotted
    if PT_plot_mode:
        return pressures, temperatures
    abundances = {}
    msum = 0.0
    for species in pRT_object.line_species:
        abundances[species] = 10**parameters[species.split("_R_")[0]].value * np.ones_like(pressures)
        msum += 10**parameters[species.split("_R_")[0]].value
    abundances['H2'] = 0.766 * (1.0-msum) * np.ones_like(pressures)
    abundances['He'] = 0.234 * (1.0-msum) * np.ones_like(pressures)

    MMW = calc_MMW(abundances)
    for cloud in pRT_object.cloud_species:
        cname = cloud.split('_')[0]
        pbase = Pbases[cname]
        abundances[cname] = np.zeros_like(temperatures)
        msum += 10**parameters['log_X_cb_'+cname].value
        try:
            abundances[cname][pressures < pbase] = \
                            10**parameters['log_X_cb_'+cname].value *\
                            ((pressures[pressures <= pbase]/pbase)**parameters['fsed'].value)
        except:
            return None,None

    # This is bad for some reason...
    #if msum > 1.0:
    #    return None, None
    gravity = 2.0
    R_pl = 1.0
    if 'log_g' in parameters.keys() and 'mass' in parameters.keys():
        gravity = 10**parameters['log_g'].value
        R_pl = np.sqrt(nc.G*parameters['mass'].value/gravity)
    elif 'log_g' in parameters.keys():
        gravity= 10**parameters['log_g'].value
        R_pl = parameters['R_pl'].value
    elif 'mass' in parameters.keys():
        R_pl = parameters['R_pl'].value
        gravity = nc.G * parameters['mass'].value/R_pl**2
    else:
        print("Pick two of log_g, R_pl and mass priors!")
        sys.exit(5)
    print(np.log10(gravity), R_pl/nc.r_jup)
    pRT_object.calc_flux(temperatures, \
                     abundances, \
                     gravity, \
                     MMW, \
                     contribution = False,
                     fsed = parameters['fsed'].value,
                     Kzz = 10**parameters['log_kzz'].value * np.ones_like(pressures),
                     sigma_lnorm = parameters['sigma_lnorm'].value)

    wlen_model = nc.c/pRT_object.freq/1e-4
    wlen = nc.c/pRT_object.freq
    f_lambda = pRT_object.flux*nc.c/wlen**2.
    # convert to flux per m^2 (from flux per cm^2) cancels with step below
    #f_lambda = f_lambda * 1e4
    # convert to flux per micron (from flux per cm) cancels with step above
    #f_lambda = f_lambda * 1e-4
    # convert from ergs to Joule
    f_lambda = f_lambda * 1e-7
    spectrum_model = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)
    return wlen_model, spectrum_model

def guillot_eqchem_transmission(pRT_object, \
                                    parameters, \
                                    PT_plot_mode = False,
                                    AMR = False):
    """
    Equilibrium Chemistry Transmission Model, Guillot Profile

    This model computes a transmission spectrum based on equilibrium chemistry
    and a Guillot temperature-pressure profile.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  Rstar : Radius of the host star [cm]
                *  log_g : Log of surface gravity
                *  R_pl : planet radius [cm]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  gamma : Guillot gamma parameter
                *  log_kappa_IR : The log of the ratio between the infrared and optical opacities
                *  Fe/H : Metallicity
                *  C/O : Carbon to oxygen ratio
                *  Pcloud : optional, cloud base pressure of a grey cloud deck.
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
    """

    try:
        from petitRADTRANS import poor_mans_nonequ_chem as pm
    except ImportError:
        print("Could not import poor_mans_nonequ_chemistry. Exiting.")
        sys.exit(2)
    # Make the P-T profile
    pressures = pRT_object.press/1e6
    temperatures = nc.guillot_global(pressures, \
                                    10**parameters['log_kappa_IR'].value, \
                                    parameters['gamma'].value, \
                                    10**parameters['log_g'].value, \
                                    parameters['Tint'].value, \
                                    parameters['Tequ'].value)

    # If in evaluation mode, and PTs are supposed to be plotted
    if PT_plot_mode:
        return pressures, temperatures

    # Make the abundance profile
    COs = parameters['C/O'].value * np.ones_like(pressures)
    FeHs = parameters['[Fe/H]'].value * np.ones_like(pressures)

    abundances_interp = pm.interpol_abundances(COs, \
                                               FeHs, \
                                               temperatures, \
                                               pressures)

    abundances = {}
    for species in pRT_object.line_species:
        abundances[species] = abundances_interp[ \
                    species.split('_R_')[0].replace('_all_iso', '').replace('C2H2','C2H2,acetylene')]
    abundances['H2'] = abundances_interp['H2']
    abundances['He'] = abundances_interp['He']

    MMW = abundances_interp['MMW']

    # Calculate the spectrum
    if 'Pcloud' in parameters.keys():
        pRT_object.calc_transm(temperatures, \
                               abundances, \
                               10**parameters['log_g'].value, \
                               MMW, \
                               R_pl=parameters['R_pl'].value, \
                               P0_bar=0.01,
                               Pcloud = parameters['Pcloud'].value)
    else:
        pRT_object.calc_transm(temperatures, \
                               abundances, \
                               10**parameters['log_g'].value, \
                               MMW, \
                               R_pl=parameters['R_pl'].value, \
                               P0_bar=0.01)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model = (pRT_object.transm_rad/parameters['Rstar'].value)**2.

    return wlen_model, spectrum_model

def isothermal_eqchem_transmission(pRT_object, \
                                    parameters, \
                                    PT_plot_mode = False,
                                    AMR = False):
    """
    Equilibrium Chemistry Transmission Model, Isothermal Profile

    This model computes a transmission spectrum based on equilibrium chemistry
    and a Guillot temperature-pressure profile.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  Rstar : Radius of the host star [cm]
                *  log_g : Log of surface gravity
                *  R_pl : planet radius [cm]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  Fe/H : Metallicity
                *  C/O : Carbon to oxygen ratio
                *  Pcloud : optional, cloud base pressure of a grey cloud deck.
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
    """

    try:
        from petitRADTRANS import poor_mans_nonequ_chem as pm
    except ImportError:
        print("Could not import poor_mans_nonequ_chemistry. Exiting.")
        sys.exit(2)
    # Make the P-T profile
    pressures = pRT_object.press/1e6
    temperatures = parameters['Temp'].value * np.ones_like(pressures)

    # If in evaluation mode, and PTs are supposed to be plotted
    if PT_plot_mode:
        return pressures, temperatures

    # Make the abundance profile
    COs = parameters['C/O'].value * np.ones_like(pressures)
    FeHs = parameters['[Fe/H]'].value * np.ones_like(pressures)

    abundances_interp = pm.interpol_abundances(COs, \
                                               FeHs, \
                                               temperatures, \
                                               pressures)

    abundances = {}
    for species in pRT_object.line_species:
        abundances[species] = abundances_interp[ \
                    species.split('_R_')[0].replace('_all_iso', '').replace('C2H2','C2H2,acetylene')]
    abundances['H2'] = abundances_interp['H2']
    abundances['He'] = abundances_interp['He']

    MMW = abundances_interp['MMW']
    pcloud = None
    if 'log_Pcloud' in parameters.keys():
        pcloud = 10**parameters['log_Pcloud'].value
    elif 'Pcloud' in parameters.keys():
        pcloud = parameters['log_Pcloud'].value
    if pcloud is not None:
        # P0_bar is important for low gravity transmission
        # spectrum. 100 is standard, 0.01 is good for small,
        # low gravity objects
        pRT_object.calc_transm(temperatures, \
                               abundances, \
                               10**parameters['log_g'].value, \
                               MMW, \
                               R_pl=parameters['R_pl'].value, \
                               P0_bar=0.01,
                               Pcloud = pcloud)
    else:
        pRT_object.calc_transm(temperatures, \
                               abundances, \
                               10**parameters['log_g'].value, \
                               MMW, \
                               R_pl=parameters['R_pl'].value, \
                               P0_bar=0.01)
    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    return wlen_model, spectrum_model


def isothermal_free_transmission(pRT_object, \
                                parameters, \
                                PT_plot_mode = False,
                                AMR = False):
    """
    Free Chemistry Transmission Model, Guillot Profile

    This model computes a transmission spectrum based on free retrieval chemistry
    and an isothermal temperature-pressure profile.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  Rstar : Radius of the host star [cm]
                *  log_g : Log of surface gravity
                *  R_pl : planet radius [cm]
                *  Temp : Isothermal temperature [K]
                *  species : Abundances for each species used in the retrieval
                *  Pcloud : optional, cloud base pressure of a grey cloud deck.
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
    """

    # Make the P-T profile
    pressures = pRT_object.press/1e6
    temperatures = parameters['Temp'].value * np.ones_like(pressures)
    #for key,value in parameters.items():
    #    print(key,value.value)
    # If in evaluation mode, and PTs are supposed to be plotted
    if PT_plot_mode:
        return pressures, temperatures
    abundances = {}
    msum = 0.0
    for species in pRT_object.line_species:
        spec = species.split('_R_')[0]
        abundances[species] = 10**parameters[spec].value * np.ones_like(pressures)
        msum += 10**parameters[spec].value
    if msum > 1.0:
        return None, None
    abundances['H2'] = 0.766 * (1.0-msum) * np.ones_like(pressures)
    abundances['He'] = 0.234 * (1.0-msum) * np.ones_like(pressures)

    #MMW = abundances_interp['MMW']
    MMW = calc_MMW(abundances)

    # Calculate the spectrum
    pcloud = 100.0
    if 'Pcloud' in parameters.keys():
        pcloud = parameters['Pcloud'].value
    elif 'log_Pcloud' in parameters.keys():
        pcloud = 10**parameters['log_Pcloud'].value
    # Calculate the spectrum

    if pcloud is not None:
        # P0_bar is important for low gravity transmission
        # spectrum. 100 is standard, 0.01 is good for small,
        # low gravity objects
        pRT_object.calc_transm(temperatures, \
                               abundances, \
                               10**parameters['log_g'].value, \
                               MMW, \
                               R_pl=parameters['R_pl'].value, \
                               P0_bar=0.01,
                               Pcloud = pcloud)
    else:
        pRT_object.calc_transm(temperatures, \
                               abundances, \
                               10**parameters['log_g'].value, \
                               MMW, \
                               R_pl=parameters['R_pl'].value, \
                               P0_bar=0.01)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    return wlen_model, spectrum_model


##################
# Helper Functions
##################

### Global Guillot P-T formula with kappa/grav replaced by delta
def PT_ret_model(T3, delta, alpha, tint, press, FeH, CO, conv = True):
    '''
    Self-luminous retrieval P-T model.

    Arsg:
        T3 : np.array([t1, t2, t3])
            temperature points to be added on top
            radiative Eddington structure (above tau = 0.1).
            Use spline interpolation, t1 < t2 < t3 < tconnect as prior.
        delta :
            proportionality factor in tau = delta * press_cgs**alpha
        alpha:
            power law index in tau = delta * press_cgs**alpha
            For the tau model: use proximity to kappa_rosseland photosphere
            as prior.
        tint:
            internal temperature of the Eddington model
        press:
            input pressure profile in bar
        conv:
            enforce convective adiabat yes/no
        CO:
            C/O for the nabla_ad interpolation
        FeH:
            metallicity for the nabla_ad interpolation
    '''

    try:
        from petitRADTRANS import poor_mans_nonequ_chem as pm
    except ImportError:
        print("Could not import poor_mans_nonequ_chemistry. Exiting.")
        sys.exit(2)
    # Go grom bar to cgs
    press_cgs = press*1e6

    # Calculate the optical depth
    tau = delta*press_cgs**alpha

    # This is the eddington temperature
    tedd = (3./4.*tint**4.*(2./3.+tau))**0.25

    ab = pm.interpol_abundances(CO*np.ones_like(tedd), \
            FeH*np.ones_like(tedd), \
            tedd, \
            press)

    nabla_ad = ab['nabla_ad']

    # Enforce convective adiabat
    if conv:
        # Calculate the current, radiative temperature gradient
        nab_rad = np.diff(np.log(tedd))/np.diff(np.log(press_cgs))
        # Extend to array of same length as pressure structure
        nabla_rad = np.ones_like(tedd)
        nabla_rad[0] = nab_rad[0]
        nabla_rad[-1] = nab_rad[-1]
        nabla_rad[1:-1] = (nab_rad[1:]+nab_rad[:-1])/2.

        # Where is the atmosphere convectively unstable?
        conv_index = nabla_rad > nabla_ad

        # TODO: Check remains convective and convergence
        for i in range(10):
            if i == 0:
                t_take = cp.copy(tedd)
            else:
                t_take = cp.copy(tfinal)

            ab = pm.interpol_abundances(CO*np.ones_like(t_take), \
                FeH*np.ones_like(t_take), \
                t_take, \
                press)

            nabla_ad = ab['nabla_ad']

            # Calculate the average nabla_ad between the layers
            nabla_ad_mean = nabla_ad
            nabla_ad_mean[1:] = (nabla_ad[1:]+nabla_ad[:-1])/2.
            # What are the increments in temperature due to convection
            tnew = nabla_ad_mean[conv_index]*np.mean(np.diff(np.log(press_cgs)))
            # What is the last radiative temperature?
            tstart = np.log(t_take[~conv_index][-1])
            # Integrate and translate to temperature from log(temperature)
            tnew = np.exp(np.cumsum(tnew)+tstart)

            # Add upper radiative and
            # lower conective part into one single array
            tfinal = cp.copy(t_take)
            tfinal[conv_index] = tnew

            if np.max(np.abs(t_take-tfinal)/t_take) < 0.01:
                #print('n_ad', 1./(1.-nabla_ad[conv_index]))
                break

    else:
        tfinal = tedd

    # Add the three temperature-point P-T description above tau = 0.1
    def press_tau(tau):
        # Returns the pressure at a given tau, in cgs
        return (tau/delta)**(1./alpha)

    # Where is the uppermost pressure of the Eddington radiative structure?
    p_bot_spline = press_tau(0.1)

    for i_intp in range(2):

        if i_intp == 0:

            # Create the pressure coordinates for the spline support nodes at low pressure
            support_points_low = np.logspace(np.log10(press_cgs[0]), \
                             np.log10(p_bot_spline), \
                             4)

            # Create the pressure coordinates for the spline support nodes at high pressure,
            # the corresponding temperatures for these nodes will be taken from the
            # radiative+convective solution
            support_points_high = 1e1**np.arange(np.log10(p_bot_spline),
                                                 np.log10(press_cgs[-1]),
                                                 np.diff(np.log10(support_points_low))[0])

            # Combine into one support node array, don't add the p_bot_spline point twice.
            support_points = np.zeros(len(support_points_low)+len(support_points_high)-1)
            support_points[:4] = support_points_low
            support_points[4:] = support_points_high[1:]

        else:

            # Create the pressure coordinates for the spline support nodes at low pressure
            support_points_low = np.logspace(np.log10(press_cgs[0]), \
                             np.log10(p_bot_spline), \
                             7)

            # Create the pressure coordinates for the spline support nodes at high pressure,
            # the corresponding temperatures for these nodes will be taken from the
            # radiative+convective solution
            support_points_high = np.logspace(np.log10(p_bot_spline), np.log10(press_cgs[-1]), 7)

            # Combine into one support node array, don't add the p_bot_spline point twice.
            support_points = np.zeros(len(support_points_low)+len(support_points_high)-1)
            support_points[:7] = support_points_low
            support_points[7:] = support_points_high[1:]

        # Define the temperature values at the node points.
        t_support = np.zeros_like(support_points)

        if i_intp == 0:
            tfintp = interp1d(press_cgs, tfinal,kind='cubic')
            # The temperature at p_bot_spline (from the radiative-convectice solution)
            t_support[int(len(support_points_low))-1] = tfintp(p_bot_spline)
            # The temperature at pressures below p_bot_spline (free parameters)
            t_support[:(int(len(support_points_low))-1)] = T3
            # t_support[:3] = tfintp(support_points_low)
            # The temperature at pressures above p_bot_spline
            # (from the radiative-convectice solution)
            t_support[int(len(support_points_low)):] = \
                tfintp(support_points[(int(len(support_points_low))):])

        else:
            tfintp1 = interp1d(press_cgs, tret,kind='cubic')
            t_support[:(int(len(support_points_low))-1)] = \
                tfintp1(support_points[:(int(len(support_points_low))-1)])

            tfintp = interp1d(press_cgs, tfinal)
            # The temperature at p_bot_spline (from the radiative-convectice solution)
            t_support[int(len(support_points_low))-1] = tfintp(p_bot_spline)
            #print('diff', t_connect_calc - tfintp(p_bot_spline))
            t_support[int(len(support_points_low)):] = \
                tfintp(support_points[(int(len(support_points_low))):])

        # Make the temperature spline interpolation to be returned to the user
        cs = CubicSpline(np.log10(support_points), t_support)
        tret = cs(np.log10(press_cgs))

    tret[tret<0.0] = 10.0
    # Return the temperature, the pressure at tau = 1,
    # and the temperature at the connection point.
    # The last two are needed for the priors on the P-T profile.
    return tret#, press_tau(1.)/1e6, tfintp(p_bot_spline)

def _make_half_pressure_better(P_clouds, press):
    """
    deprecated
    """
    # Obsolete, replaced with fixed_length_amr
    press_plus_index = np.zeros((press.shape[0],2))
    press_plus_index[:,0] = press
    press_plus_index[:,1] = range(len(press))

    press_small = press_plus_index[::24, :]
    press_plus_index = press_plus_index[::2,:]

    indexes_small = press_small[:,0] > 0.
    indexes       = press_plus_index[:,0] > 0.

    for cname,P_cloud in P_clouds.items():
        indexes_small = indexes_small & \
            ((np.log10(press_small[:,0]/P_cloud) > 0.05) | \
            (np.log10(press_small[:,0]/P_cloud) < -0.3))
        indexes = indexes & \
            ((np.log10(press_plus_index[:,0]/P_cloud) > 0.05) | \
            (np.log10(press_plus_index[:,0]/P_cloud) < -0.3))

    press_cut = press_plus_index[~indexes, :]
    press_small_cut = press_small[indexes_small, :]

    press_out = np.zeros((len(press_cut)+len(press_small_cut))*2).reshape((len(press_cut)+len(press_small_cut)), 2)
    press_out[:len(press_small_cut), :] = press_small_cut
    press_out[len(press_small_cut):, :] = press_cut

    press_out = np.sort(press_out, axis = 0)
    return press_out[:,0],  press_out[:, 1].astype('int')

def fixed_length_amr(P_clouds, press, scaling = 10, width = 3):
    """
    This function takes in the cloud base pressures for each cloud,
    and returns an array of pressures with a high resolution mesh
    in the region where the cloud is located.

    Args:
        P_clouds : numpy.ndarray
            The cloud base pressures in bar
        press : np.ndarray
            The high resolution pressure array.
        scaling : int
            The factor by which the low resolution pressure array is scaled
        width : int
            The number of low resolution bins to be replaced for each cloud layer.
    """

    # P_clouds is array of pressures
    # press should be ~len scaling*100
    # guarantees total length will be press.shape[0] + P_clouds.shape[0]*width*(scaling - 1)
    # scaling is how many hi-res points per normal point. Must be int
    # width is the number of low res points to replace. Must be int.

    press_plus_index = np.zeros((press.shape[0],2))
    press_plus_index[:,0] = np.logspace(np.log10(np.min(press)),np.log10(np.max(press)),press.shape[0])
    press_plus_index[:,1] = range(len(press_plus_index[:,0]))
    # Set up arrays for indexing
    press_small= press_plus_index[::scaling,:]
    # Make some lists to store the replacement indices
    c_list = []
    for i,P_cloud in enumerate(P_clouds):
        # Find out where the clouds are in the high res grid
        idx = (np.abs(press_plus_index[:,0] - P_cloud)).argmin()
        # constant length list of indices around that point
        inds = np.linspace(int(idx-(width*scaling/2.0)),int(idx+(width*scaling/2.0))-1,int(scaling*width),dtype=int)
        c_list.append(inds)
    # We need to return a list that's always the same length
    # So we need to check for duplicates
    total_inds = []
    for j in range(len(c_list)):
        # At first, just copy in the list
        if j == 0:
            total_inds.extend(c_list[j])
            continue
        # Check if the next set of indices is lower than the current minimum
        # if so, we want to add scaling*width indices below the current minimum
        if min(c_list[j])<=min(np.array(total_inds)):
            start = c_list[j][-1]
            sl = len(total_inds)
            ind = 0
            done = 0
            while len(total_inds) < sl+int(scaling*width):
                if start-ind < 0:
                    start = start + len(c_list[j])
                    ind = done
                if np.in1d(start-ind,np.array(total_inds)).any():
                    ind += 1
                    continue
                else:
                    total_inds.append(int(start-ind))
                    ind+=1
                    done += 1
        # Check if the smallest new index is larger than the current max
        # if so, we can just add the indexes
        # I can probably replace all this with total_inds.extend(c_list[j])
        elif max(c_list[j])>=max(np.array(total_inds)):
            start = c_list[j][0]
            sl = len(total_inds)
            ind = 0
            done = 0
            while len(total_inds) < sl+int(scaling*width):
                if (start+ind) >= (len(press_plus_index)-1):
                    start = start - len(c_list[j]) - 1
                    ind = done
                    continue
                if np.in1d(start+ind,np.array(total_inds)).any():
                    ind += 1
                    continue
                else:
                    total_inds.append(int(start+ind))
                    ind+=1
                    done += 1
        else:
            # This loop takes care of cases where we're between existing entries
            # it adds indices until duplicates are found, then keeps incrementing
            # until there is a free index to add.
            start = np.array(total_inds)[np.where(np.array(total_inds)==c_list[j][0])[0]]
            sl = len(total_inds)
            ind = 0
            done = 0
            while len(total_inds) < sl+int(scaling*width):
                if (start+ind) >= (len(press_plus_index)-1):
                    start = start - len(c_list[j]) - 1
                    ind = done
                    continue
                if np.in1d(start+ind,np.array(total_inds)).any():
                    ind += 1
                    continue
                else:
                    total_inds.append(int(start+ind))
                    ind+=1
                    done += 1
    total_inds = np.array(sorted(total_inds,reverse=False))
    # Stack the low res and high res grids, sort it, and take the unique values
    try:
        press_out = np.vstack((press_small,press_plus_index[total_inds]))
    except:
        print("AMR returned incorrect length")
        return PGLOBAL, np.array([0])
    press_out = np.sort(press_out, axis = 0)
    p_out,ind = np.unique(press_out[:,0],return_index = True)
    return p_out,  press_out[ind, 1].astype('int')


def get_abundances(pressures, temperatures, line_species, cloud_species, parameters, AMR = False):
    """
    This function takes in the C/O ratio, metallicity, and quench pressures and uses them
    to compute the gas phase and equilibrium condensate abundances.
    Clouds are currently hard coded into the function.

    Args:
        pressures : numpy.ndarray
            A log spaced pressure array. If AMR is on it should be the full high resolution grid.
        temperatures : numpy.ndarray
            A temperature array with the same shape as pressures
        line_species : List(str)
            A list of gas species that will contribute to the line-by-line opacity of the pRT atmosphere.
        cloud_species : List(str)
            A list of condensate species that will contribute to the cloud opacity of the pRT atmosphere.
        parameters : dict
            A dictionary of model parameters, in particular it must contain the names C/O, Fe/H and
            log_pquench. Additionally the cloud parameters log_X_cb_Fe(c) and MgSiO3(c) must be present.
        AMR : bool
            Turn the adaptive mesh grid on or off. See fixed_length_amr for implementation.

    Returns:
        abundances : dict
            Mass fraction abundances of all atmospheric species
        MMW : numpy.ndarray
            Array of the mean molecular weights in each pressure bin
        small_index : numpy.ndarray
            The indices of the high resolution grid to use to define the adaptive grid.
    """

    try:
        from petitRADTRANS import poor_mans_nonequ_chem as pm
    except ImportError:
        print("Could not import poor_mans_nonequ_chemistry. Exiting.")
        sys.exit(2)
    # Make the abundance profile
    COs = parameters['C/O'].value * np.ones_like(pressures)
    FeHs = parameters['Fe/H'].value * np.ones_like(pressures)

    # Prior check all input params
    XFe = fc.return_XFe(parameters['Fe/H'].value, parameters['C/O'].value)
    XMgSiO3 = fc.return_XMgSiO3(parameters['Fe/H'].value, parameters['C/O'].value)

    clouds = {}
    clouds['Fe(c)'] = 10**parameters['log_X_cb_Fe(c)'].value*XFe
    clouds['MgSiO3(c)'] = 10**parameters['log_X_cb_MgSiO3(c)'].value*XMgSiO3

    pquench_C = None
    if 'log_pquench' in parameters.keys():
        pquench_C = 10**parameters['log_pquench'].value
    abundances_interp = pm.interpol_abundances(COs, \
                                               FeHs, \
                                               temperatures, \
                                               pressures,
                                               Pquench_carbon = pquench_C)
    MMW = abundances_interp['MMW']

    Pbases = {}
    Pbases['Fe(c)'] = fc.simple_cdf_Fe(pressures, temperatures,
                                parameters['Fe/H'].value, parameters['C/O'].value, np.mean(MMW))
    Pbases['MgSiO3(c)'] = fc.simple_cdf_MgSiO3(pressures, temperatures,
                                parameters['Fe/H'].value, parameters['C/O'].value, np.mean(MMW))
    #Pbases['KCL(c)'] = fc.simple_cdf_KCL(pressures, temperatures,
    #                            parameters['Fe/H'].value, parameters['C/O'].value, np.mean(MMW))
    #Pbases['Na2S(c)'] = fc.simple_cdf_Na2S(pressures, temperatures,
    #                            parameters['Fe/H'].value, parameters['C/O'].value, np.mean(MMW))
    fseds = {}
    abundances = {}
    # Clouds
    p_clouds = []
    for key, val in Pbases.items():
        p_clouds.append(val)
    p_clouds = np.array(p_clouds)

    if AMR:
        press_use, small_index = fixed_length_amr(p_clouds,
                                                  pressures,
                                                  parameters['pressure_scaling'].value,
                                                  parameters['pressure_width'].value)
    else :
        #TODO: Test
        press_use = pressures
        small_index = np.indices(press_use)

    for cloud in cp.copy(cloud_species):
        cname = cloud.split('_')[0]
        if 'log_X_cb_'+cname not in parameters.keys():
            continue
        abundances[cname] = np.zeros_like(temperatures)
        abundances[cname][pressures < Pbases[cname]] = \
                        clouds[cname] *\
                        (pressures[pressures <= Pbases[cname]]/\
                        Pbases[cname])**parameters['fsed'].value
        fseds[cname] = parameters['fsed'].value

    if AMR:
        abundances['Fe(c)'] = abundances['Fe(c)'][small_index]
        abundances['MgSiO3(c)'] = abundances['MgSiO3(c)'][small_index]
        for species in line_species:
            abundances[species] = abundances_interp[species.split('_')[0]][small_index]
        abundances['H2'] = abundances_interp['H2'][small_index]
        abundances['He'] = abundances_interp['He'][small_index]

    else:
        for species in line_species:
            abundances[species] = abundances_interp[species.split('_')[0]]
        abundances['H2'] = abundances_interp['H2']
        abundances['He'] = abundances_interp['He']
    if 'FeH' in abundances.keys():
        abundances['FeH'] = abundances['FeH']/2.
    return abundances,MMW,small_index

def pglobal_check(press,shape,scaling):
    """
    Check to ensure that the global pressure array has the correct length.
    Updates PGLOBAL.

    Args:
        press : numpy.ndarray
            Pressure array from a pRT_object. Used to set the min and max values of PGLOBAL
        shape : int
            the shape of the pressure array if no AMR is used
        scaling :
            The factor by which the pressure array resolution should be scaled.
    """

    global PGLOBAL
    if PGLOBAL.shape[0] != int(scaling*shape):
        PGLOBAL = np.logspace(np.log10(press[0]),
                                np.log10(press[1]),
                                int(scaling*shape))

def set_resolution(lines,abundances,resolution):
    """
    deprecated
    """
    # Set correct key names in abundances for pRT, with set resolution
    # Only needed for free chemistry retrieval
    #print(lines)
    #print(abundances)
    if resolution is None:
        return abundances
    for line in lines:
        abundances[line] = abundances[line.split("_R_"+str(resolution))[0]]
        del abundances[line.split("_R_"+str(resolution))]
    return abundances



"""
This module contains a set of useful functions that don't really fit anywhere
else. This includes flux conversions, prior functions, mean molecular weight
calculations, transforms from mass to number fractions, and fits file output.
"""
import sys
import os
# To not have numpy start parallelizing on its own
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
from scipy.special import gamma,erfcinv
import numpy as np
import math as math
from molmass import Formula
#import threading, subprocess


SQRT2 = math.sqrt(2.)

#################
# Flux scaling
#################
def surf_to_meas(flux,p_rad,dist):
    """
    surf_to_meas
    Convert from emission flux to measured flux at earth
    Args:
        flux : numpy.ndarray
            Absolute flux value or spectrum as emitted by a source of radius p_rad
        p_rad : float
            Planet radius, in same units as dist
        dist : float
            Distance to the object, in the same units as p_rad
    Returns:
        m_flux : numpy.ndarray
            Apparent flux
    """

    m_flux = flux * p_rad**2/dist**2
    return m_flux


#################
# Prior Functions
#################
# Stolen from https://github.com/JohannesBuchner/MultiNest/blob/master/src/priors.f90
def log_prior(cube,lx1,lx2):
    return 10**(lx1+cube*(lx2-lx1))

def uniform_prior(cube,x1,x2):
    return x1+cube*(x2-x1)

def gaussian_prior(cube,mu,sigma):
    return mu + sigma*SQRT2*erfcinv(2.0*(1.0 - cube))
    #return -(((cube-mu)/sigma)**2.)/2.

def log_gaussian_prior(cube,mu,sigma):
    bracket = sigma*sigma + sigma*SQRT2*erfcinv(2.0*cube)
    return mu*np.exp(bracket)

def delta_prior(cube,x1,x2):
    return x1

# Sanity checks on parameter ranges
def b_range(x, b):
    if x > b:
        return -np.inf
    else:
        return 0.

def a_b_range(x, a, b):
    if x < a:
        return -np.inf
    elif x > b:
        return -np.inf
    else:
        return 0.


########################
# Mean Molecular Weights
########################

def getMM(species):
    """
    Get the molecular mass of a given species.

    This function uses the molmass package to
    calculate the mass number for the standard
    isotope of an input species. If all_iso
    is part of the input, it will return the
    mean molar mass.

    Args:
        species : string
            The chemical formula of the compound. ie C2H2 or H2O
    Returns:
        The molar mass of the compound in atomic mass units.
    """
    name = species.split("_")[0]
    name = name.split(',')[0]
    f = Formula(name)
    if "all_iso" in species:
        return f.mass
    return f.isotope.massnumber

def calc_MMW(abundances):
    """
    calc_MMW
    Calculate the mean molecular weight in each layer.

    Args:
        abundances : dict
            dictionary of abundance arrays, each array must have the shape of the pressure array used in pRT,
            and contain the abundance at each layer in the atmosphere.
    """
    MMW = 0.
    for key in abundances.keys():
        # exo_k resolution
        spec = key.split("_R_")[0]
        MMW += abundances[key]/getMM(spec)
    return 1./MMW

def get_MMW_from_mfrac(m_frac):
    """
    wraps calc_MMW
    """

    calc_MMW(m_frac)

def get_MMW_from_nfrac(n_frac):
    """
    Calculate the mean molecular weight from a number fraction

    Args:
        n_fracs : dict
            A dictionary of number fractions
    """

    mass = 0.0
    for key,value in n_frac.items():
        spec = key.split("_R_")[0]
        mass += value*getMM(spec)
    return mass

def mass_to_number(m_frac):
    """
    Convert mass fractions to number fractions

    Args:
        m_fracs : dict
            A dictionary of mass fractions
    """

    n_frac = {}
    MMW = get_MMW_from_mfrac(m_frac)
    for key,value in m_frac.items():
        spec = key.split("_R_")[0]
        n_frac[key]=value/getMM(spec)*MMW
    return n_frac

def number_to_mass(n_fracs):
    """
    Convert number fractions to mass fractions

    Args:
        n_fracs : dict
            A dictionary of number fractions
    """

    m_frac = {}
    MMW = get_MMW_from_nfrac(n_fracs)
    for key,value in n_fracs.items():
        spec = key.split("_R_")[0]
        m_frac[key] = value*getMM(spec)/MMW
    return m_frac

########################
# File Formatting
########################
def fits_output(wavelength, spectrum, covariance, object, output_dir = "", correlation = None):
    """
    Generate a fits file that can be used as an input to a pRT retrieval.

    Args:
        wavelength : numpy.ndarray
            The wavelength bin centers in micron. dim(N)
        spectrum : numpy.ndarray
            The flux density in W/m2/micron at each wavelength bin. dim(N)
        covariance : numpy.ndarray
            The covariance of the flux in (W/m2/micron)^2 dim(N,N)
        object : string
            The name of the object, used for file naming.
        output_dir : string
            The parent directory of the output file.
        correlation : numpy.ndarray
            The correlation matrix of the flux points (See Brogi & Line 2018, https://arxiv.org/pdf/1811.01681.pdf)

    Returns:
        hdul : astropy.fits.HDUlist
            The HDUlist object storing the spectrum.
    """

    from astropy.io import fits
    primary_hdu = fits.PrimaryHDU([])
    primary_hdu.header['OBJECT'] = object
    c1 = fits.Column(name = "WAVELENGTH", array = wavelength, format = 'D',unit = "micron")
    c2 = fits.Column(name = "FLUX", array = spectrum, format = 'D',unit = "W/m2/micron")
    c3 = fits.Column(name = "COVARIANCE", array = covariance, format = str(covariance.shape[0])+'D',unit = "[W/m2/micron]^2")
    if correlation is not None:
        c4 = fits.Column(name = "CORRELATION", array = correlation, format =  str(correlation.shape[0])+'D',unit = " - ")
    columns = [c1,c2,c3,c4]
    table_hdu = fits.BinTableHDU.from_columns(columns,name = 'SPECTRUM')
    hdul = fits.HDUList([primary_hdu,table_hdu])
    outstring = os.path.join(output_dir, object + "_spectrum.fits")
    hdul.writeto(outstring,overwrite=True,checksum=True,output_verify='exception')
    return hdul



import numpy as np
import copy as cp
from scipy.interpolate import interp1d

plotting = False
if plotting:
    import pylab as plt
    from petitRADTRANS import nat_cst as nc
    
#############################################################
# Cloud Cond
#############################################################
# This file allows the calculation of equilibrium cloud abundances
# and base pressures
#
# TODO: Make a better cloud module.

#############################################################
# To calculate X_Fe from [Fe/H], C/O
#############################################################

# metal species
metals = ['C','N','O','Na','Mg','Al','Si','P','S','Cl','K','Ca','Ti','V','Fe','Ni']

# solar abundances, [Fe/H] = 0, from Asplund+ 2009
nfracs = {}
nfracs['H'] = 0.9207539305
nfracs['He'] = 0.0783688694
nfracs['C'] = 0.0002478241
nfracs['N'] = 6.22506056949881e-05
nfracs['O'] = 0.0004509658
nfracs['Na'] = 1.60008694353205e-06
nfracs['Mg'] = 3.66558742055362e-05
nfracs['Al'] = 2.595e-06
nfracs['Si'] = 2.9795e-05
nfracs['P'] = 2.36670201997668e-07
nfracs['S'] = 1.2137900734604e-05
nfracs['Cl'] = 2.91167958499589e-07
nfracs['K'] = 9.86605611925677e-08
nfracs['Ca'] = 2.01439011429255e-06
nfracs['Ti'] = 8.20622804366359e-08
nfracs['V'] = 7.83688694089992e-09
nfracs['Fe'] = 2.91167958499589e-05
nfracs['Ni'] = 1.52807116806281e-06

# atomic masses
masses = {}
masses['H'] = 1.
masses['He'] = 4.
masses['C'] = 12.
masses['N'] = 14.
masses['O'] = 16.
masses['Na'] = 23.
masses['Mg'] = 24.3
masses['Al'] = 27.
masses['Si'] = 28.
masses['P'] = 31.
masses['S'] = 32.
masses['Cl'] = 35.45
masses['K'] = 39.1
masses['Ca'] = 40.
masses['Ti'] = 47.9
masses['V'] = 51.
masses['Fe'] = 55.8
masses['Ni'] = 58.7

def return_XFe(FeH, CO):

    nfracs_use = cp.copy(nfracs)

    for spec in nfracs.keys():

        if (spec != 'H') and (spec != 'He'):
            nfracs_use[spec] = nfracs[spec]*1e1**FeH

    nfracs_use['O'] = nfracs_use['C']/CO

    XFe = masses['Fe']*nfracs_use['Fe']
    add = 0.
    for spec in nfracs_use.keys():
        add += masses[spec]*nfracs_use[spec]

    XFe = XFe / add

    return XFe

def return_XMgSiO3(FeH, CO):

    nfracs_use = cp.copy(nfracs)

    for spec in nfracs.keys():

        if (spec != 'H') and (spec != 'He'):
            nfracs_use[spec] = nfracs[spec]*1e1**FeH

    nfracs_use['O'] = nfracs_use['C']/CO

    nfracs_mgsio3 = np.min([nfracs_use['Mg'], \
                            nfracs_use['Si'], \
                            nfracs_use['O']/3.])
    masses_mgsio3 = masses['Mg'] \
      + masses['Si'] \
      + 3. * masses['O']
      
    Xmgsio3 = masses_mgsio3*nfracs_mgsio3
    add = 0.
    for spec in nfracs_use.keys():
        add += masses[spec]*nfracs_use[spec]

    Xmgsio3 = Xmgsio3 / add

    return Xmgsio3

def return_XNa2S(FeH, CO):

    nfracs_use = cp.copy(nfracs)

    for spec in nfracs.keys():

        if (spec != 'H') and (spec != 'He'):
            nfracs_use[spec] = nfracs[spec]*1e1**FeH

    nfracs_use['O'] = nfracs_use['C']/CO

    nfracs_na2s = np.min([nfracs_use['Na']/2., \
                            nfracs_use['S']])
    masses_na2s = 2.*masses['Na'] \
      + masses['S']
      
    Xna2s = masses_na2s*nfracs_na2s
    add = 0.
    for spec in nfracs_use.keys():
        add += masses[spec]*nfracs_use[spec]

    Xna2s = Xna2s / add

    return Xna2s

def return_XKCL(FeH, CO):

    nfracs_use = cp.copy(nfracs)

    for spec in nfracs.keys():

        if (spec != 'H') and (spec != 'He'):
            nfracs_use[spec] = nfracs[spec]*1e1**FeH

    nfracs_use['O'] = nfracs_use['C']/CO

    nfracs_kcl = np.min([nfracs_use['K'], \
                            nfracs_use['Cl']])
    masses_kcl = masses['K'] \
      + masses['Cl']
      
    Xkcl = masses_kcl*nfracs_kcl
    add = 0.
    for spec in nfracs_use.keys():
        add += masses[spec]*nfracs_use[spec]

    Xkcl = Xkcl / add

    return Xkcl

#############################################################
# Fe saturation pressure, from Ackerman & Marley (2001), including erratum (P_vap is in bar, not cgs!)
#############################################################

def return_T_cond_Fe(FeH, CO, MMW = 2.33):

    T = np.linspace(100.,10000.,1000)
    # Taken from Ackerman & Marley (2001)
    # including their erratum
    P_vap = lambda x: np.exp(15.71 - 47664./x)

    XFe = return_XFe(FeH, CO)

    return P_vap(T)/(XFe*MMW/masses['Fe']), T

def return_T_cond_Fe_l(FeH, CO, MMW = 2.33):

    T = np.linspace(100.,10000.,1000)
    # Taken from Ackerman & Marley (2001)
    # including their erratum
    P_vap = lambda x: np.exp(9.86 - 37120./x)

    XFe = return_XFe(FeH, CO)

    return P_vap(T)/(XFe*MMW/masses['Fe']), T

def return_T_cond_Fe_comb(FeH, CO, MMW = 2.33):

    P1, T1 = return_T_cond_Fe(FeH, CO, MMW)
    P2, T2 = return_T_cond_Fe_l(FeH, CO, MMW)

    retP = np.zeros_like(P1)
    index = P1<P2
    retP[index] = P1[index]
    retP[~index] = P2[~index]
    return retP, T2

def return_T_cond_Fe_free(XFe, MMW = 2.33):

    T = np.linspace(100.,10000.,1000)
    # Taken from Ackerman & Marley (2001)
    # including their erratum
    P_vap = lambda x: np.exp(15.71 - 47664./x)
    return P_vap(T)/(XFe*MMW/masses['Fe']), T

def return_T_cond_Fe_l_free(XFe, MMW = 2.33):

    T = np.linspace(100.,10000.,1000)
    # Taken from Ackerman & Marley (2001)
    # including their erratum
    P_vap = lambda x: np.exp(9.86 - 37120./x)
    return P_vap(T)/(XFe*MMW/masses['Fe']), T

def return_T_cond_Fe_comb_free(XFe, MMW = 2.33):

    P1, T1 = return_T_cond_Fe_free(XFe, MMW)
    P2, T2 = return_T_cond_Fe_l_free(XFe, MMW)
    retP = np.zeros_like(P1)
    index = P1<P2
    retP[index] = P1[index]
    retP[~index] = P2[~index]
    return retP, T2

def return_T_cond_MgSiO3(FeH, CO, MMW = 2.33):

    T = np.linspace(100.,10000.,1000)
    # Taken from Ackerman & Marley (2001)
    # including their erratum
    P_vap = lambda x: np.exp(25.37 - 58663./x)

    Xmgsio3 = return_XMgSiO3(FeH, CO)

    m_mgsio3 =  masses['Mg'] \
      + masses['Si'] \
      + 3. * masses['O']
    return P_vap(T)/(Xmgsio3*MMW/m_mgsio3), T
def return_T_cond_MgSiO3_free(Xmgsio3, MMW = 2.33):

    T = np.linspace(100.,10000.,1000)
    # Taken from Ackerman & Marley (2001)
    # including their erratum
    P_vap = lambda x: np.exp(25.37 - 58663./x)
    m_mgsio3 =  masses['Mg'] \
      + masses['Si'] \
      + 3. * masses['O']
    return P_vap(T)/(Xmgsio3*MMW/m_mgsio3), T

def return_T_cond_Na2S(FeH, CO, MMW = 2.33):

    # Taken from Charnay+2018
    T = np.linspace(100.,10000.,1000)
    # This is the partial pressure of Na, so
    # Divide by factor 2 to get the partial
    # pressure of the hypothetical Na2S gas
    # particles, this is OK: there are
    # more S than Na atoms at solar
    # abundance ratios.
    P_vap = lambda x: 1e1**(8.55 - 13889./x - 0.5*FeH)/2.

    Xna2s = return_XNa2S(FeH, CO)

    m_na2s =  2.*masses['Na'] \
      + masses['S']
    return P_vap(T)/(Xna2s*MMW/m_na2s), T
def return_T_cond_Na2S_free(Xna2s, MMW = 2.33):

    # Taken from Charnay+2018
    T = np.linspace(100.,10000.,1000)
    # This is the partial pressure of Na, so
    # Divide by factor 2 to get the partial
    # pressure of the hypothetical Na2S gas
    # particles, this is OK: there are
    # more S than Na atoms at solar
    # abundance ratios.
    P_vap = lambda x: 1e1**(8.55 - 13889./x - 0.5*FeH)/2.
    m_na2s =  2.*masses['Na'] \
      + masses['S']
    return P_vap(T)/(Xna2s*MMW/m_na2s), T

def return_T_cond_KCL(FeH, CO, MMW = 2.33):

    # Taken from Charnay+2018
    T = np.linspace(100.,10000.,1000)
    P_vap = lambda x: 1e1**(7.611 - 11382./T)

    Xkcl = return_XKCL(FeH, CO)

    m_kcl =  masses['K'] \
      + masses['Cl']
    return P_vap(T)/(Xkcl*MMW/m_kcl), T
def return_T_cond_KCL_free(Xkcl, MMW = 2.33):

    # Taken from Charnay+2018
    T = np.linspace(100.,10000.,1000)
    P_vap = lambda x: 1e1**(7.611 - 11382./T)
    m_kcl =  masses['K'] \
      + masses['Cl']
    return P_vap(T)/(Xkcl*MMW/m_kcl), T


if plotting:
    
    #FeHs = np.linspace(-0.5, 2., 5)
    #COs = np.linspace(0.3, 1.2, 5)
    FeHs = [0.]
    COs = [0.55]

    for FeH in FeHs:
        for CO in COs:
            P, T = return_T_cond_Fe(FeH, CO)
            plt.plot(T,P, label = 'Fe(c), [Fe/H] = '+str(FeH)+', C/O = '+str(CO), color = 'black')
            P, T = return_T_cond_Fe_l(FeH, CO)
            plt.plot(T,P, '--', label = 'Fe(l), [Fe/H] = '+str(FeH)+', C/O = '+str(CO))
            P, T = return_T_cond_Fe_comb(FeH, CO)
            plt.plot(T,P, ':', label = 'Fe(c+l), [Fe/H] = '+str(FeH)+', C/O = '+str(CO))
            P, T = return_T_cond_MgSiO3(FeH, CO)
            plt.plot(T,P, label = 'MgSiO3, [Fe/H] = '+str(FeH)+', C/O = '+str(CO))
            P, T = return_T_cond_Na2S(FeH, CO)
            plt.plot(T,P, label = 'Na2S, [Fe/H] = '+str(FeH)+', C/O = '+str(CO))
            P, T = return_T_cond_KCL(FeH, CO)
            plt.plot(T,P, label = 'KCL, [Fe/H] = '+str(FeH)+', C/O = '+str(CO))
            

    plt.yscale('log')
    '''
    plt.xlim([0., 5000.])
    plt.ylim([1e5,1e-10])
    '''
    plt.xlim([0., 2000.])
    plt.ylim([1e2,1e-3])
    plt.legend(loc = 'best', frameon = False)
    plt.show()

def simple_cdf_Fe(press, temp, FeH, CO, MMW = 2.33):

    Pc, Tc = return_T_cond_Fe_comb(FeH, CO, MMW)
    index = (Pc > 1e-8) & (Pc < 1e5)
    Pc, Tc = Pc[index], Tc[index]
    tcond_p = interp1d(Pc, Tc)
    #print(Pc, press)
    Tcond_on_input_grid = tcond_p(press)

    Tdiff = Tcond_on_input_grid - temp
    diff_vec = Tdiff[1:]*Tdiff[:-1]
    ind_cdf = (diff_vec < 0.)
    if len(diff_vec[ind_cdf]) > 0:
        P_clouds = (press[1:]+press[:-1])[ind_cdf]/2.
        P_cloud = P_clouds[-1]
    else:
        P_cloud = 1e-8

    if plotting:
        plt.plot(temp, press)
        plt.plot(Tcond_on_input_grid, press)
        plt.axhline(P_cloud, color = 'red', linestyle = '--')
        plt.yscale('log')
        plt.xlim([0., 3000.])
        plt.ylim([1e2,1e-6])
        plt.show()

    return P_cloud
def simple_cdf_Fe_free(press, temp, XFe, MMW = 2.33):

    Pc, Tc = return_T_cond_Fe_comb_free(XFe, MMW)
    index = (Pc > 1e-8) & (Pc < 1e5)
    Pc, Tc = Pc[index], Tc[index]
    tcond_p = interp1d(Pc, Tc)
    #print(Pc, press)
    Tcond_on_input_grid = tcond_p(press)

    Tdiff = Tcond_on_input_grid - temp
    diff_vec = Tdiff[1:]*Tdiff[:-1]
    ind_cdf = (diff_vec < 0.)
    if len(diff_vec[ind_cdf]) > 0:
        P_clouds = (press[1:]+press[:-1])[ind_cdf]/2.
        P_cloud = P_clouds[-1]
    else:
        P_cloud = 1e-8

    if plotting:
        plt.plot(temp, press)
        plt.plot(Tcond_on_input_grid, press)
        plt.axhline(P_cloud, color = 'red', linestyle = '--')
        plt.yscale('log')
        plt.xlim([0., 3000.])
        plt.ylim([1e2,1e-6])
        plt.show()

    return P_cloud
def simple_cdf_MgSiO3(press, temp, FeH, CO, MMW = 2.33):

    Pc, Tc = return_T_cond_MgSiO3(FeH, CO, MMW)
    index = (Pc > 1e-8) & (Pc < 1e5)
    Pc, Tc = Pc[index], Tc[index]
    tcond_p = interp1d(Pc, Tc)
    #print(Pc, press)
    Tcond_on_input_grid = tcond_p(press)

    Tdiff = Tcond_on_input_grid - temp
    diff_vec = Tdiff[1:]*Tdiff[:-1]
    ind_cdf = (diff_vec < 0.)
    if len(diff_vec[ind_cdf]) > 0:
        P_clouds = (press[1:]+press[:-1])[ind_cdf]/2.
        P_cloud = P_clouds[-1]
    else:
        P_cloud = 1e-8

    if plotting:
        plt.plot(temp, press)
        plt.plot(Tcond_on_input_grid, press)
        plt.axhline(P_cloud, color = 'red', linestyle = '--')
        plt.yscale('log')
        plt.xlim([0., 3000.])
        plt.ylim([1e2,1e-6])
        plt.show()

    return P_cloud
def simple_cdf_MgSiO3_free(press, temp, Xmgsio3, MMW = 2.33):

    Pc, Tc = return_T_cond_MgSiO3_free(Xmgsio3, MMW)
    index = (Pc > 1e-8) & (Pc < 1e5)
    Pc, Tc = Pc[index], Tc[index]
    tcond_p = interp1d(Pc, Tc)
    #print(Pc, press)
    Tcond_on_input_grid = tcond_p(press)

    Tdiff = Tcond_on_input_grid - temp
    diff_vec = Tdiff[1:]*Tdiff[:-1]
    ind_cdf = (diff_vec < 0.)
    if len(diff_vec[ind_cdf]) > 0:
        P_clouds = (press[1:]+press[:-1])[ind_cdf]/2.
        P_cloud = P_clouds[-1]
    else:
        P_cloud = 1e-8

    if plotting:
        plt.plot(temp, press)
        plt.plot(Tcond_on_input_grid, press)
        plt.axhline(P_cloud, color = 'red', linestyle = '--')
        plt.yscale('log')
        plt.xlim([0., 3000.])
        plt.ylim([1e2,1e-6])
        plt.show()

    return P_cloud
def simple_cdf_Na2S(press, temp, FeH, CO, MMW = 2.33):

    Pc, Tc = return_T_cond_Na2S(FeH, CO, MMW)
    index = (Pc > 1e-8) & (Pc < 1e5)
    Pc, Tc = Pc[index], Tc[index]
    tcond_p = interp1d(Pc, Tc)
    #print(Pc, press)
    Tcond_on_input_grid = tcond_p(press)

    Tdiff = Tcond_on_input_grid - temp
    diff_vec = Tdiff[1:]*Tdiff[:-1]
    ind_cdf = (diff_vec < 0.)
    if len(diff_vec[ind_cdf]) > 0:
        P_clouds = (press[1:]+press[:-1])[ind_cdf]/2.
        P_cloud = P_clouds[-1]
    else:
        P_cloud = 1e-8

    if plotting:
        plt.plot(temp, press)
        plt.plot(Tcond_on_input_grid, press)
        plt.axhline(P_cloud, color = 'red', linestyle = '--')
        plt.yscale('log')
        plt.xlim([0., 3000.])
        plt.ylim([1e2,1e-6])
        plt.show()

    return P_cloud

def simple_cdf_KCL(press, temp, FeH, CO, MMW = 2.33):

    Pc, Tc = return_T_cond_KCL(FeH, CO, MMW)
    index = (Pc > 1e-8) & (Pc < 1e5)
    Pc, Tc = Pc[index], Tc[index]
    tcond_p = interp1d(Pc, Tc)
    #print(Pc, press)
    Tcond_on_input_grid = tcond_p(press)

    Tdiff = Tcond_on_input_grid - temp
    diff_vec = Tdiff[1:]*Tdiff[:-1]
    ind_cdf = (diff_vec < 0.)
    if len(diff_vec[ind_cdf]) > 0:
        P_clouds = (press[1:]+press[:-1])[ind_cdf]/2.
        P_cloud = P_clouds[-1]
    else:
        P_cloud = 1e-8

    if plotting:
        plt.plot(temp, press)
        plt.plot(Tcond_on_input_grid, press)
        plt.axhline(P_cloud, color = 'red', linestyle = '--')
        plt.yscale('log')
        plt.xlim([0., 3000.])
        plt.ylim([1e2,1e-6])
        plt.show()

    return P_cloud

if plotting:
    kappa_IR = 0.01
    gamma = 0.4
    T_int = 200.
    T_equ = 1550.
    gravity = 1e1**2.45

    pressures = np.logspace(-6, 2, 100)

    temperature = nc.guillot_global(pressures, kappa_IR, gamma, gravity, T_int, T_equ)

    simple_cdf_Fe(pressures, temperature, 0., 0.55)
    simple_cdf_MgSiO3(pressures, temperature, 0., 0.55)

    T_int = 200.
    T_equ = 800.
    temperature = nc.guillot_global(pressures, kappa_IR, gamma, gravity, T_int, T_equ)
    simple_cdf_Na2S(pressures, temperature, 0., 0.55)

    T_int = 150.
    T_equ = 650.
    temperature = nc.guillot_global(pressures, kappa_IR, gamma, gravity, T_int, T_equ)
    simple_cdf_KCL(pressures, temperature, 0., 0.55)
    
